/*
 * Formats.java
 *
 * Version: $Revision: 1545 $
 *
 * Date: $Date: 2006-07-10 16:04:02 -0400 (Mon, 10 Jul 2006) $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.administer;

import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;

import org.dspace.authorize.AuthorizeException;
import org.dspace.content.Bitstream;
import org.dspace.content.BitstreamFormat;
import org.dspace.content.Bundle;
import org.dspace.content.DSpaceObject;
import org.dspace.content.Item;
import org.dspace.content.format.FormatIdentifierManager;
import org.dspace.content.format.FormatHit;
import org.dspace.content.format.FormatConfidence;
import org.dspace.content.format.FormatRegistryManager;
import org.dspace.content.format.FormatRegistryException;
import org.dspace.core.Constants;
import org.dspace.core.Context;
import org.dspace.handle.HandleManager;

/**
 * Formats -- Workbench for format registries and format identification
 * <pre>
 * The primary functions are:
 *  1. Generate a report on the formats of selected Bitstreams, and
 *     optionally on the outcome of attempting to re-guess their formats
 *
 *  2. Generate Histogram report of usage of BitstreamFormats and Registries
 *     by Bitstreams in the archive.  Can also include details about
 *     confidence or source.
 *
 *  3. Re-guess, interactively, the format for a set of Bitstreams:
 *
 *  4. Update (change) bitstream's format to a designated one.
 *
 *  5. List all known BitstreamFormats in the local cache.
 *
 *  Run the command with --help for further details.
 * </pre>
 * @author Larry Stone
 * @version $Revision: 1545 $
 */
public class Formats
{
    // column-separator character for output,  set with -F option (like awk)
    private static char sepChar = ',';

    // maps Confidence to a compact numeric index -- actual Confidence values are sparse.
    private static Map<FormatConfidence,Integer> confidence2index  = new HashMap<FormatConfidence,Integer>();
    private static int confidence2indexLength = 0;

    private static Options options = new Options();

    static
    {
        int i = 0;
        for (FormatConfidence fc : FormatConfidence.values())
        {
            confidence2index.put(fc, new Integer(i++));
        }
        confidence2indexLength = i;
    }

    private static void warn(String msg)
    {
        System.err.println(msg+"\n");
    }

    private static void die(String msg, boolean usage)
    {
        if (msg != null)
            System.err.println("ERROR: "+msg+"\n");
        if (usage)
        {
            HelpFormatter myhelp = new HelpFormatter();
            System.out.println("===== DSpace BitstreamFormat Identification Workbench =====");
            myhelp.printHelp("\njava "+Formats.class.getName()+" [ -R | -H | -G | -L ] [options..]\n", options);
            System.out.println("\nChoose one primary function: -R, -G, -H, -L, or -C.");
        }
        System.exit(1);
    }

    public static void main(String[] argv) throws Exception
    {
        CommandLineParser parser = new PosixParser();
        options.addOption("h", "help", false, "Show this help message.");

        // Options for primary function, must be mutually-exclusive
        OptionGroup og = new OptionGroup();
        og.addOption(new Option("R", "report", false, "Generate report on Format, Confidence, Source on each selected Bitstream."));
        og.addOption(new Option("G", "guess", false, "Interactively Guess (identify) Bitstreams' format"));
        og.addOption(new Option("H", "histogram", false, "Show Histogram of BitstreamFormat and Format Registry usage"));
        og.addOption(new Option("L", "list", false, "List all BitstreamFormat (cache) entries in the DB by number, name, etc."));
        og.addOption(OptionBuilder.withLongOpt("change")
                      .hasArg(true).withArgName("BSF DB ID")
                      .withDescription("forcibly change Bitstreams' Format to the one identified by this DB ID.")
                      .create("C"));
        og.setRequired(true);
        options.addOptionGroup(og);

        // Options for the "report" primary operation
        options.addOption("g", "guess-report", false, "add to -R report, what re-guessing would do.");

        // Options for the "guess" primary operation
        options.addOption("y", "yes", false, "automatically choose first option when guessing (identifying) new format.");

        // options on report, guess, change
        options.addOption("a", "all", false, "Operate on  ALL bitstreams (after filters applied)");
        options.addOption(OptionBuilder.withLongOpt("item")
                      .hasArgs().withArgName("Item DB ID")
                      .withDescription("Operate on Bitstream(s) in this Item (value is Handle or databaseID)")
                      .create("i"));
        options.addOption(OptionBuilder.withLongOpt("sequence")
                      .hasArg(true).withArgName("SequenceID")
                      .withDescription("selects Bitstream matching this SequenceID within the chosen Item")
                      .create("s"));
        options.addOption(OptionBuilder.withLongOpt("bitstream")
                      .hasArgs().withArgName("Bitstream DB ID")
                      .withDescription("Operate Bitstream matching this DB ID (instead of -i & -s)")
                      .create("b"));
        options.addOption(OptionBuilder.withLongOpt("format-id")
                      .hasArg(true).withArgName("BSF DB ID")
                      .withDescription("filter: operate only on Bitstreams whose BitstreamFormat matches this DB ID.")
                      .create("f"));
        options.addOption(OptionBuilder.withLongOpt("source")
                      .hasArg(true).withArgName("Regular Exp")
                      .withDescription("filter: operate only on selected Bitstreams whose Format Source also match this regular expression.")
                      .create("S"));
        options.addOption("u", "ufd", false, "filter: only operate on Bitstreams that have a UserFormatDescription set.");
        options.addOption("v", "verbose", false, "show verbose details on Report and Guess.");
        options.addOption("n", "dryrun", false, "do not actually make any changes to state of archive.");

        // options on histogram reports
        options.addOption(OptionBuilder.withLongOpt("columns")
                      .hasArg(true).withArgName("comma-sep letters")
                      .withDescription("Columns to display in Histogram report: C=confidence, M=MIMEtype,S=source,E=external-ID")
                      .create("c"));

        // field separator in some histogram output lines
        options.addOption(OptionBuilder.withLongOpt("field-sep")
                      .hasArg(true).withArgName("character")
                      .withDescription("field separator character in output lines (e.g. \",\")")
                      .create("F"));

        CommandLine line = null;
        try
        {
            line = parser.parse(options, argv);
        }
        catch (ParseException e)
        {
            die("Error in command options: "+e, true);
        }

        if (line.hasOption('h'))
            die(null, true);
        Context c = new Context();
        c.setIgnoreAuthorization(true);
        try
        {
            // set field separator
            if (line.hasOption('F'))
            {
                String ss = line.getOptionValue('F');
                if (ss.length() != 1)
                   die("-F field separator value must be one character!", true);
                sepChar = ss.charAt(0);
            }

            if (line.hasOption('H'))
                new Formats().doHistogram(c, line.getOptionValue('c'));

            else if (line.hasOption('L'))
                new Formats().doList(c, line.hasOption('v'));

            else if (line.hasOption('G'))
                (new GuessAction(line)).main(c);

            else if (line.hasOption('R'))
                (new ReportAction(line)).main(c);

            else if (line.hasOption('C'))
                (new ChangeAction(line)).main(c);

            else
                die ("Missing required option to select subcommand, must include one of -R, -H, -G.", true);
        }
        catch (Exception e)
        {
            System.err.println("Error:  " + e.toString());
            e.printStackTrace();
        }
    }

    // resolve an arg referring to an Item by ID or Handle
    private static Item resolveArg(Context c, String arg)
            throws SQLException, AuthorizeException
    {
        DSpaceObject result = null;

        // if there's a "/" assume it's a handle
        if (arg.indexOf('/') != -1)
        {
            result = HandleManager.resolveToObject(c, arg);
            if (result == null ||
                result.getType() != Constants.ITEM)
                result = null;
        }
        else
        {
            result = Item.find(c, Integer.parseInt(arg));
        }
        return (Item)result;
    }

    // Driver for all commands that operate on set of Bitstreams selected
    // by the -a or -i options, with filters -s, -u, -S, -f etc.
    // Contains the logic that parses command and selects Bitstreams.
    static abstract class FormatsAction
    {
        protected CommandLine options = null;

        protected boolean verbose = false;

        protected boolean dryrun = false;

        protected FormatsAction(CommandLine options)
        {
            super();
            this.options = options;
            this.verbose = options.hasOption('v');
            this.dryrun = options.hasOption('n');
        }

        // process a single Bitstream
        abstract void handleBitstream(Context context, Bitstream bs, Item item)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException;

        // perform action of entire command; should call processBitstreams()
        // this wrapper method allows setup before and summary after the loop.
        abstract void main(Context c)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException;

        // call the implementation's handleBitstream() on selected Bitstreams.
        // returns exit status, 0 = success.
        protected int processBitstreams(Context c)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException
        {
            int status = 0;
            boolean all = options.hasOption('a');
            int formatFilter = -1;
            if (options.hasOption('f'))
            {
                formatFilter = Integer.parseInt(options.getOptionValue('f'));
                // sanity check
                if (BitstreamFormat.find(c,formatFilter) == null)
                    die("No such BitstreamFormat with ID="+options.getOptionValue('f'), false);
            }
            String sourceFilter = options.getOptionValue('S');
            boolean ufdFilter = options.hasOption('u');
                                      
            // All bitstreams, filtered by Source, Format, and/or SequenceID
            if (all)
            {
                Iterator<Bitstream> bi = Bitstream.findAll(c);
                while (bi.hasNext())
                {
                    Bitstream bs = bi.next();
                    if ((formatFilter < 0 ||
                         formatFilter == bs.getFormat().getID()) &&
                        (!ufdFilter || bs.getUserFormatDescription() != null) &&
                        (sourceFilter == null ||
                         bs.getFormatSource().matches(sourceFilter)))
                        handleBitstream(c, bs, null);
                    c.removeCached(bs, bs.getID());
                }
            }
             
            // Items' bitstreams, filtered by Source, Format, and/or SequenceID
            else if (options.hasOption('i'))
            {
                int seq = -1;
                String sArg = options.getOptionValue('s');
                if (sArg != null)
                    seq = Integer.parseInt(sArg);

                for (String iArg : options.getOptionValues('i'))
                {
                    Item item = resolveArg(c, iArg);
                    if (item == null)
                    {
                        warn("SKIPPING: Failed to resolve item: "+iArg);
                        status = 1;
                    }
                    else
                    {
                      ITEM:
                        for (Bundle bn : item.getBundles())
                        {
                            for (Bitstream bbs : bn.getBitstreams())
                            {
                                if (seq < 0 &&
                                    (formatFilter < 0 ||
                                     formatFilter == bbs.getFormat().getID()) &&
                                    (!ufdFilter ||
                                     bbs.getUserFormatDescription() != null) &&
                                    (sourceFilter == null ||
                                     bbs.getFormatSource().matches(sourceFilter)))
                                {
                                    handleBitstream(c, bbs, item);
                                }
                                else
                                {
                                     if (bbs.getSequenceID() == seq)
                                     {
                                        handleBitstream(c, bbs, item);
                                        break ITEM;
                                     }
                                }
                            }
                        }
                    }
                }
            }

            // selected single bitstream(s)
            else if (options.hasOption('b'))
            {
                for (String bArg : options.getOptionValues('b'))
                {
                    try
                    {
                        Bitstream bs = Bitstream.find(c, Integer.parseInt(bArg));
                        if (bs == null)
                        {
                            warn("Could not find bitstream DB ID="+bArg+", SKIPPING.");
                            status = 1;
                        }
                        else
                            handleBitstream(c, bs, null);
                    }
                    catch (NumberFormatException e)
                    {
                        warn("Skipping Illegal Bitstream DB ID: "+bArg+": "+e);
                        status = 1;
                    }
                }
            }
             
            // sanity check, no args.
            else
                die("Missing options to select Bitstreams to process.", true);

            return status;
        }
    }

    // Interactively Guess formats on selected Bitstreams
    static class GuessAction extends FormatsAction
    {
        boolean anySuccess = false;
        boolean yes = false;

        protected GuessAction(CommandLine options)
        {
            super(options);
        }

        void main(Context c)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException
        {
            anySuccess = false;
            yes = options.hasOption('y');

            int status = processBitstreams(c);

            if (anySuccess)
                c.complete();
            System.exit(status);
        }

        void handleBitstream(Context c, Bitstream bs, Item item)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException
        {
            String desc = "";
            if (item != null)
                desc = " from Item "+item.getHandle()+", SequenceID="+bs.getSequenceID();
             
            String ufd = bs.getUserFormatDescription();
            System.out.println("Guessing format of Bitstream ID="+bs.getID()+","+desc+"\n  named "+bs.getName()+
                ",\n  Old "+
                ((ufd == null) ? "format="+bs.getFormat().getName()+", " :
                                "UserDefinedFormat=\""+ufd+"\",\n  ")+
                "conf="+bs.getFormatConfidence()+", source=\""+bs.getFormatSource()+"\"");
            if (yes)
            {
                FormatHit h = FormatIdentifierManager.identifyFormat(c, bs);
                System.out.println("Set it to format "+
                    FormatRegistryManager.makeIdentifier(h.namespace,h.identifier)+
                    ", Conf="+h.confidence+", src="+h.source);
                if (!dryrun)
                {
                    h.applyToBitstream(c, bs);
                    anySuccess = true;
                }
            }
            else
            {
                List<FormatHit> hits = FormatIdentifierManager
                    .identifyAllFormats(c, bs);
                System.out.println("\nChoose from "+hits.size()+" Hits..");
                    System.out.println(
                      String.format("%1$3s %2$-20.20s %3$-16.16s %4$-13.13s %5$-23.23s",
                        "-#-", "-name-", "-external-ID-", "-confidence-", "-source-"));
                int i = 1;
                for (FormatHit h : hits)
                {
                    String source = (h.source != null && h.source.length() > 21) ?
                                    (".."+h.source.substring(h.source.length()-21)) : h.source;
                    System.out.println(
                      String.format("%1$3d %2$-20.20s %3$-16.16s %4$-13.13s %5$-23.23s",
                        i,
                        h.name,
                        h.getIdentifier(),
                        h.confidence,
                        source));
                    if (h.warning != null)
                        System.out.println("  Warning: "+h.warning);
                    ++i;
                }
                System.out.println("\nEnter number of hit to choose or 0 to skip:");
                Reader ir = new InputStreamReader(System.in);
                char cline[] = new char[80];
                int len = ir.read(cline, 0, 80);
                if (len <= 0)
                    return;
                String sline = new String(cline, 0, len).trim();
                if (sline.length() > 0)
                {
                    int choice = Integer.parseInt(sline);
                    if (choice > 0 && choice <= hits.size())
                    {
                        FormatHit h = hits.get(choice-1);
                        System.out.println("Set it to format "+
                          FormatRegistryManager.makeIdentifier(h.namespace, h.identifier));
                        if (!dryrun)
                        {
                            h.applyToBitstream(c, bs);
                            anySuccess = true;
                        }
                        return;
                    }
                }
            }
        }
    }

    // Generate detail and summary reports on selected Bitstreams
    static class ReportAction extends FormatsAction
    {
        // table constants: stats for report
        private static final int totalSeen            = 0;
        private static final int sameFormatSameConf   = 1;
        private static final int sameFormatBetterConf = 2;
        private static final int sameFormatWorseConf  = 3;
        private static final int diffFormat           = 4;
        private static final int unknownBefore        = 5;
        private static final int unknownAfter         = 6;
        private static final int unknownBoth          = 7;
         
        // Must keep labels in sync with symbols above
        private static String guessLabel[] =
            {
                "Total Bitstreams",
                "Same Format/Same Confidence",
                "Same Format/Better Confidence",
                "Same Format/ Worse Confidence",
                "Different Format",
                "Unknown Format Before/Now Identified",
                "Unknown Format After/Had been Identified",
                "Unknown Format both Before and After",
            };
         
        private int report[] = null;

        private String reportFmt = null;
         
        private Map bsPerSourcePerConfidence[] = null;

        private boolean doGuess = false;

        protected ReportAction(CommandLine options)
        {
            super(options);
        }

        void main(Context c)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException
        {
            doGuess = options.hasOption('g');
            report = new int[8];
            Arrays.fill(report, 0);
             
            // set up bitstream counter data structure
            int maxBsfID = -1;
            for (BitstreamFormat bsf : BitstreamFormat.findAll(c))
                maxBsfID = Math.max(maxBsfID, bsf.getID());
            bsPerSourcePerConfidence = new Map[maxBsfID+1];
            for (BitstreamFormat bsf : BitstreamFormat.findAll(c))
                bsPerSourcePerConfidence[bsf.getID()] =
                    new HashMap<String,int[]>();
             
            if (verbose)
            {
                String header = "# BITSTREAM%1$c <Handle>%1$c <DBID>%1$c <Name>%1$c <Current-Format>%1$c <current-Confidence>%1$c <current-Source>";
                reportFmt = "BITSTREAM%1$c %2$s%1$c %3$5d%1$c \"%4$s\"%1$c %5$4d%1$c %6$s%1$c \"%7$s\"";
                if (doGuess)
                {
                    header += "%1$c <primary-result>%1$c <secondary-result>%1$c <new-Format>%1$c <new-Confidence>%1$c <new-Source>";
                    reportFmt += "%1$c %8$d%1$c %9$d%1$c %10$4d%1$c %11$s%1$c \"%12$s\"%1$c";
                }
                System.out.println(String.format(header, sepChar));
            }
             
            processBitstreams(c);
             
            // print format count summary
            System.out.println(String.format("# FORMAT-SUMMARY%1$c <bsf-id>%1$c <bsf-name>%1$c <count>%1$c <source>%1$c <confidence>", sepChar));
            for (int i = 0; i < bsPerSourcePerConfidence.length; ++i)
            {
                if (bsPerSourcePerConfidence[i] != null)
                {
                    Map<String,int[]> countPerSourcePerConfidence = (Map<String,int[]>)bsPerSourcePerConfidence[i];
                    for (Map.Entry<String,int[]> ce : countPerSourcePerConfidence.entrySet())
                    {
                        int countPerConfidence[] = ce.getValue();
                        for (FormatConfidence fc : FormatConfidence.values())
                        {
                            int j = confidence2index.get(fc).intValue();
                            if (countPerConfidence[j] > 0)
                                System.out.println(
                                  String.format("FORMAT-SUMMARY%1$c %2$4d%1$c \"%3$s\"%1$c %4$5d%1$c \"%5$s\"%1$c %6$s", sepChar,
                                    i,
                                    BitstreamFormat.find(c,i).getName(),
                                    countPerConfidence[j],
                                    ce.getKey(),
                                    fc));
                        }
                    }
                }
            }
            // print format guess summary
            if (doGuess)
            {
                System.out.println(String.format("# GUESS-SUMMARY%1$c <result-type>%1$c <count>%1$c <description>", sepChar));
                for (int i = 0; i < guessLabel.length; ++i)
                    System.out.println(String.format("GUESS-SUMMARY%1$c %2$-2d%1$c %3$6d%1$c \"%4$s\"",
                                                     sepChar, i, report[i], guessLabel[i]));
            }
        }

        void handleBitstream(Context c, Bitstream bs, Item item)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException
        {
            ++report[totalSeen];
            int primary = -1, secondary = -1;
            FormatHit h = null;
            BitstreamFormat newBsf = null;
            int fmtID = bs.getFormat().getID();
            FormatConfidence fmtConfidence = bs.getFormatConfidence();
            String fmtSource = bs.getFormatSource();
             
            // accumulate stats
            Map<String,int[]> countPerSourcePerConfidence =
                (Map<String,int[]>) bsPerSourcePerConfidence[bs.getFormat().getID()];
            if (countPerSourcePerConfidence.containsKey(fmtSource))
                ++(countPerSourcePerConfidence.get(fmtSource))[confidence2index.get(fmtConfidence)];
            else
            {
                int counts[] = new int[confidence2indexLength];
                Arrays.fill(counts, 0);
                ++counts[confidence2index.get(fmtConfidence)];
                countPerSourcePerConfidence.put(fmtSource, counts);
            }
             
            if (doGuess)
            {
                h = FormatIdentifierManager.identifyFormat(c, bs);
                newBsf = BitstreamFormat
                    .findByIdentifier(c, h.namespace, h.identifier, false);
                if (bs.getFormat().isUnknown())
                {
                    if (h.confidence == FormatConfidence.UNIDENTIFIED)
                    {
                        secondary = unknownBoth;
                        FormatConfidence oldfc = bs.getFormatConfidence();
                        int cmp = oldfc == null ? -1 : oldfc.compareTo(h.confidence);
                        if (cmp == 0)
                            primary = sameFormatSameConf;
                        else if (cmp < 0)
                            primary = sameFormatBetterConf;
                        else
                            primary = sameFormatWorseConf;
                    }
                    else
                    {
                        secondary = unknownBefore;
                        primary = diffFormat;
                    }
                }
                else if (h.confidence == FormatConfidence.UNIDENTIFIED)
                {
                    secondary = unknownAfter;
                    primary = diffFormat;
                }
                else
                {
                    if (newBsf != null && newBsf.equals(bs.getFormat()))
                    {
                        FormatConfidence oldfc = bs.getFormatConfidence();
                        int cmp = oldfc == null ? -1 : oldfc.compareTo(h.confidence);
                        if (cmp == 0)
                            primary = sameFormatSameConf;
                        else if (cmp < 0)
                            primary = sameFormatBetterConf;
                        else
                            primary = sameFormatWorseConf;
                    }
                    else
                        primary = diffFormat;
                }
                ++report[primary];
                if (secondary >= 0)
                    ++report[secondary];
            }
            if (verbose)
            {
                String hdl = (item == null)? null : item.getHandle() + "/" + String.valueOf(bs.getSequenceID());
                System.out.println(
                  String.format(reportFmt, sepChar,
                     hdl, bs.getID(), bs.getName(), fmtID, fmtConfidence, fmtSource,
                     primary, secondary,
                     (newBsf == null ? null : newBsf.getID()),
                     (h == null ? null : h.confidence),
                     (h == null ? null : h.source)));
            }
        }
    }

    // display histogram of BSF and Registry usage; always operates on ALL Bitstreams
    private void doHistogram(Context c, String columns)
        throws SQLException, AuthorizeException, FormatRegistryException, IOException
    {
        boolean showMIME = false;
        boolean showExternal = false;
        boolean showConfidence = false;
        boolean showSource = false;

        if (columns != null)
        {
            for (String col : columns.split("\\s*,\\s*"))
            {
                col = col.toLowerCase();
                if (col.startsWith("m"))
                    showMIME = true;
                else if (col.startsWith("e"))
                    showExternal = true;
                else if (col.startsWith("c"))
                    showConfidence = true;
                else if (col.startsWith("s"))
                    showSource = true;
            }
        }
        if (showConfidence && showSource)
            die ("Not allowed to choose both Source and Confidence in Histogram report!", false);

        // per-registry counts
        String registries[] = FormatRegistryManager.getAllNamespaces(c);
        Map<String,Integer> registry2index = new HashMap<String,Integer>();
        for (int i = 0; i < registries.length; ++i)
            registry2index.put(registries[i], new Integer(i));
        int bsPerRegistry[] = new int[registries.length];
        Arrays.fill(bsPerRegistry, 0);
        int bsfPerRegistry[] = new int[registries.length];
        Arrays.fill(bsfPerRegistry, 0);

        // per-BSF counts
        int unknownID = BitstreamFormat.findUnknown(c).getID();
        int bsWithUfd = 0;
        int maxBsfID = -1;
        for (BitstreamFormat bsf : BitstreamFormat.findAll(c))
        {
            maxBsfID = Math.max(maxBsfID, bsf.getID());

            // update registry counts for each identifier: first get
            // normalized set of namespaces, since NS may appear multiple times:
            Set<String> namespacesPerBsf = new HashSet<String>();
            for (String nsid : bsf.getIdentifiers())
                namespacesPerBsf.add(FormatRegistryManager.namespaceOf(nsid));

            // Add ONE use for all references to one registry in each BSF.
            for (String namespace : namespacesPerBsf)
                ++bsfPerRegistry[registry2index.get(namespace).intValue()];
        }
        // "fake" BSF ID for Bitstreams with UFD
        int userDefinedFormatID = maxBsfID + 1;
        int bsPerBsf[] = new int[maxBsfID+2];
        Arrays.fill(bsPerBsf, 0);

        // XXX curse java mickeymouse generics - can't have array of typed Collection
        //Map<FormatConfidence,Integer> confidencePerBsf[] = new HashMap<FormatConfidence,Integer>[maxBsfID+2];
        Map confidencePerBsf[] = new Map[maxBsfID+2];
        if (showConfidence)
            for (int i = 0; i <= maxBsfID+1; ++i)
                confidencePerBsf[i] = new HashMap<FormatConfidence,Integer>();
        // XXX curse java mickeymouse generics
        // Map<String,Integer> sourcePerBsf[] = new HashMap<String,Integer>[maxBsfID+1];
        Map sourcePerBsf[] = new Map[maxBsfID+2];
        if (showSource)
            for (int i = 0; i <= maxBsfID+1; ++i)
                sourcePerBsf[i] = new HashMap<String,Integer>();

        boolean haveRegistry[] = new boolean[registries.length];
        Iterator<Bitstream> bi = Bitstream.findAll(c);
        while (bi.hasNext())
        {
            Bitstream bs = bi.next();
            BitstreamFormat bsf = bs.getFormat();
            int id = bsf.getID();

            // if UFD is set, use the "fake" BSF ID
            if (id == unknownID && bs.getUserFormatDescription() != null)
                id = userDefinedFormatID;

            ++bsPerBsf[id];
            if (showConfidence)
            {
                FormatConfidence fc = bs.getFormatConfidence();
                if (fc == null)
                    System.err.println("ERROR: Got NULL format confidence?  BS="+bs);
                else
                {
                    Map<FormatConfidence,Integer> cm = (Map<FormatConfidence,Integer>)confidencePerBsf[id];
                    if (cm.containsKey(fc))
                        cm.put(fc,Integer.valueOf(cm.get(fc).intValue() + 1));
                    else
                        cm.put(fc,Integer.valueOf(1));
                }
            }
            if (showSource)
            {
                String fs = bs.getFormatSource();
                Map<String,Integer> sm = (Map<String,Integer>)sourcePerBsf[id];
                if (sm.containsKey(fs))
                    sm.put(fs,Integer.valueOf(sm.get(fs).intValue() + 1));
                else
                    sm.put(fs,Integer.valueOf(1));
            }
             
            // update registry stats for each identifier
            Arrays.fill(haveRegistry, false);
            for (String nsid : bsf.getIdentifiers())
                haveRegistry[registry2index.get(FormatRegistryManager.namespaceOf(nsid)).intValue()] = true;
            for (int i = 0; i < registries.length; ++i)
            {
                if (haveRegistry[i])
                    ++bsPerRegistry[i];
            }
            c.removeCached(bs, bs.getID());
        }

        // print per-BSF stats
        // cols are: "BSF" id name bs-refs [ mime ] [ conf count ] [ src count ]
        String bsfHead = "# BSF%1$c <id>%1$c <name>%1$c <BS-ref-count>";
        String bsfFmt = "BSF%1$c %2$4d%1$c %3$-32.32s%1$c %4$8d";
        if (showExternal)
        {
            bsfFmt += "%1$c %5$-35.35s";
            bsfHead += "%1$c <External-Identifier>";
        }
        if (showMIME)
        {
            bsfFmt += "%1$c %6$-25.25s";
            bsfHead += "%1$c <MIME-type>";
        }
        if (showConfidence)
        {
            bsfFmt += "%1$c %7$-20s%1$c %8$8d";
            bsfHead += "%1$c <Confidence-type>%1$c <Confidence-count>";
        }
        else if (showSource)
        {
            bsfFmt += "%1$c %7$-35.35s%1$c %8$8d";
            bsfHead += "%1$c <Source-type>%1$c <Source-count>";
        }
        System.out.println("\n"+String.format(bsfHead, sepChar));
        boolean both[] = { false, true };
        boolean onlyFalse[] = { false };
        String ufdName = "[User Defined Format]";
        for (BitstreamFormat bsf : BitstreamFormat.findAll(c))
        {
            int id = bsf.getID();
            String name = bsf.getName();
            if (name.length() > 30)
                name = name.substring(0,30);
            name = "\""+name+"\"";

            // make an extra pass for UserFormatDescription on Unknown fmt.
            for (boolean hasUFD : (id == unknownID ? both : onlyFalse))
            {
                String extIDs[];
                if (showExternal)
                    extIDs = bsf.getIdentifiers();
                else
                {
                    extIDs = new String[1];
                    extIDs[0] = null;
                }
                // special case of UFD, access counts for bitstreams with UFD
                int printID = hasUFD ? userDefinedFormatID : id;
                String printName = hasUFD ? ufdName : name;
                for (String extID : extIDs)
                {
                    if (showConfidence)
                    {
                        Map<FormatConfidence,Integer> cm =
                          (Map<FormatConfidence,Integer>)confidencePerBsf[printID];
                        if (cm.isEmpty())
                            System.out.println(String.format(bsfFmt, sepChar,
                                id, printName, bsPerBsf[printID], extID, bsf.getMIMEType(),
                                "\"\"", 0));
                        else
                            for (Map.Entry<FormatConfidence,Integer> e : cm.entrySet())
                                System.out.println(String.format(bsfFmt, sepChar,
                                    id, printName, bsPerBsf[printID], extID, bsf.getMIMEType(),
                                    "\""+e.getKey().toString()+"\"", e.getValue().intValue()));
                    }
                    // source length is fixed to 35 in formatstring above
                    else if (showSource)
                    {
                        Map<String,Integer> sm = (Map<String,Integer>)sourcePerBsf[printID];
                        for (Map.Entry<String,Integer> e : sm.entrySet())
                        {
                            String src = e.getKey();
                            if (src == null)
                                src = "N/A";
                            else if (src.length() > 33)
                                src = src.substring(src.length()-33);
                            System.out.println(String.format(bsfFmt, sepChar,
                                id, printName, bsPerBsf[printID], extID, bsf.getMIMEType(),
                                "\""+src+"\"", e.getValue().intValue()));
                        }
                    }
                    else
                        System.out.println(String.format(bsfFmt, sepChar,
                            id, printName, bsPerBsf[printID], extID, bsf.getMIMEType()));
                }
            }
        }

        // XXX NOTE: The bitstreams-referring-to-Registry figure may seem
        // misleading, since a BS refers directly to BSF, and one BSF
        // may refer to external identifiers in *several* different
        // registries, so the BS can be counted as referring to more
        // than one registry..  Thus the sum of referring bitstreams may
        // be more than the bitstream count.  Each BS is only counted as
        // referring to a registry ONCE, however, even if it has several
        // external identifiers in that registry.

        System.out.println(String.format("\n# REGISTRY%1$c <name>%1$c <bsf-referring>%1$c <bs-referring>", sepChar));
        for (int i = 0; i < registries.length; ++i)
        {
            System.out.println(String.format("REGISTRY%1$c %2$-25.25s%1$c %3$4d%1$c%4$10d", sepChar,
                "\""+registries[i]+"\"", bsfPerRegistry[i], bsPerRegistry[i]));
        }
    }
             
    // list all BSFs for easy reference, verbose shows external-ID's too.
    private void doList(Context c, boolean showExternal)
        throws SQLException, AuthorizeException, FormatRegistryException, IOException
    {
        String bsfHead = "# BSF%1$c <id>%1$c <name>%1$c <MIMEtype>%1$c <Support>%1$c <Extension>";
        String bsfFmt = "BSF%1$c %2$4d%1$c %3$-32.32s%1$c %4$s%1$c  %5$s%1$c %6$s";
        if (showExternal)
        {
            bsfHead += "%1$c <External-Identifier>";
            bsfFmt += "%1$c %7$s";
        }
        System.out.println(String.format(bsfHead, sepChar));
        for (BitstreamFormat bsf : BitstreamFormat.findAll(c))
        {
            int id = bsf.getID();
            String name = bsf.getName();
            if (name.length() > 30)
                name = name.substring(0,30);
            name = "\""+name+"\"";

            if (showExternal)
            {
                for (String extID : bsf.getIdentifiers())
                    System.out.println(String.format(bsfFmt, sepChar,
                        id, name, bsf.getMIMEType(),
                        BitstreamFormat.supportLevelText[bsf.getSupportLevel()],
                        bsf.getCanonicalExtension(), extID));
            }
            else
                System.out.println(String.format(bsfFmt, sepChar,
                    id, name, bsf.getMIMEType(),
                    BitstreamFormat.supportLevelText[bsf.getSupportLevel()],
                    bsf.getCanonicalExtension()));
        }
    }

    // force-change the BitstreamFormat of selected BS
    static class ChangeAction extends FormatsAction
    {
        boolean anySuccess = false;
        BitstreamFormat newFormat = null;

        protected ChangeAction(CommandLine options)
        {
            super(options);
        }

        void main(Context c)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException
        {
            anySuccess = false;

            int nf = Integer.parseInt(options.getOptionValue('C'));
            newFormat = BitstreamFormat.find(c,nf);
            if (newFormat == null)
                die("No such BitstreamFormat with ID="+options.getOptionValue('C'), false);

            processBitstreams(c);

            if (anySuccess)
                c.complete();
        }

        void handleBitstream(Context c, Bitstream bs, Item item)
            throws SQLException, IOException, AuthorizeException, FormatRegistryException
        {
            if (verbose || dryrun)
            {
                String target = "Bitstream ID="+bs.getID();
                if (item != null)
                    target = "Bitstream SID="+bs.getSequenceID()+" of Item "+item.getHandle();
                System.out.println("Setting format of "+target+" to BSF ID="+newFormat.getID()+" ("+newFormat.getName()+").");
            }

            if (!dryrun)
            {
                bs.setFormat(newFormat);
                bs.setFormatSource(Formats.class.getName());
                bs.setFormatConfidence(FormatConfidence.MANUAL);
                bs.update();
                anySuccess = true;
            }
        }
    }
}
