package org.dspace.content.format;

import org.dspace.content.BitstreamFormat;
import org.dspace.content.Bitstream;
import org.dspace.core.ConfigurationManager;
import org.dspace.core.SelfNamedPlugin;
import org.dspace.core.Context;
import org.dspace.authorize.AuthorizeException;

import java.net.URL;
import java.net.MalformedURLException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Date;
import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.sql.SQLException;

import org.jdom.input.SAXBuilder;
import org.jdom.JDOMException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;

import org.apache.log4j.Logger;

/**
 * FormatRegistry implementation for locally-defined "external" format
 * registries, such as the traditional DSpace registry.  The registry
 * is loaded from a configuration file at startup (and is automatically
 * reloaded whenever that file changes).
 * <p>
 * Note that this is <i>also</i> a FormatIdentifier implementation; it
 * identifies formats by comparing filename extensions, just as the traditional
 * DSpace format registry did.
 * <p>
 * Since this is a abstract class, it must be extended by another class.
 * <p>
 * Each extension reads the following DSpace Configuration entries,
 * keyed by the value returned by getNamespace():
 * <p><pre>
 *   # path to registry contents in XML:
 *   formatRegistry.<namespace>.document = ${dspace.dir}/config/formats/NAMESPACE-formats.xml
 *   # whether or not to validate registry when loading it
 *   formatRegistry.<namespace>.validate = true
 *   # path to schema with which to validate
 *   formatRegistry.<namespace>.schema = ${dspace.dir}/config/formats/formats.xsd
 *   # Default value of supportLevel on newly created/imported BSFs,
 *   # must be one of "unknown" | "known" | "supported"
 *   formatRegistry.<namespace>.defaultSupportLevel = unknown
 * </pre><p>
 * The contents of the registry are defined by the XML document configured
 * as "document" above.  It has a simple format: The root element is
 * <registry>, it contains a series of 0 or more <entry> elements.  Each
 * entry is populated like the example shown below.
 * <p>
 * Note that the registry document is read automatically on startup, and
 * is re-interpreted if it has changed since the last read, so any changes
 * made by the system administrator take effect immediately without any need
 * to restart DSpace.
 * <p>
 * Here is an example of the XML:
 * <pre>
 * ----------------------------------------------
 *   <dsr:registry xmlns:dsr="http://www.dspace.org/xmlns/dspace/formatRegistry"
 *     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 *     xsi:schemaLocation="http://www.dspace.org/xmlns/dspace/formatRegistry http://www.dspace.org/schema/formatRegistry/1.0/formats.xsd">
 *     <dsr:entry>
 *       <dsr:name>CSS</dsr:name>
 *       <dsr:description>Cascading Style Sheets</dsr:description>
 *       <dsr:mime-type>text/css</dsr:mime-type>
 *       <dsr:external-identifiers>
 *         <dsr:identifier>DSpace:CSS</dsr:identifier>
 *         <dsr:identifier>PRONOM:x-fmt/224</dsr:identifier>
 *       </dsr:external-identifiers>
 *       <dsr:external-signatures>
 *         <!-- NOTE: all extensions must come BEFORE all filenames to pass XSD -->
 *         <!-- the first extension is assumed to be the canonical one -->
 *         <dsr:extension>css</dsr:extension>
 *         <!-- filename element only included here as an example of usage -->
 *         <dsr:filename>stylesheet</dsr:filename>
 *       </dsr:external-signatures>
 *     </dsr:entry>
 *     <!-- ...more entries... -->
 *   </dsr:registry>
 * ----------------------------------------------
 * <p>
 * @author Larry Stone
 * @version $Revision: 1323 $
 * @see FormatRegistry
 */
public abstract class BuiltinFormatRegistry
    extends SelfNamedPlugin
    implements FormatRegistry, FormatIdentifier
{
    /** record describing a builtin registry format entry */
    public class Entry
    {
        protected String name;
        protected String description = null;
        protected String mimeType = null;
        protected String extensions[] = null;
        protected String identifiers[] = null;

        // prettyprint for debugging
        public String toString()
        {
            String truncDesc = description.length() > 20 ? (description.substring(0, 20)+"..."):description;
            return "<BuiltinFormatRegistry.Entry name="+name+
                    ", description="+truncDesc+
                    ", MIME type="+mimeType+
                    ", extensions=("+joinString(extensions)+
                    "), identifiers=("+joinString(identifiers)+")>";
        }

        // make comma-separated list
        private String joinString(String sa[])
        {
            if (sa == null)
                return "";
            StringBuffer result = new StringBuffer();
            boolean first = true;
            for (String s : sa)
            {
                if (first)
                    result.append(s);
                else
                    result.append(", ").append(s);
                first = false;
            }
            return result.toString();
        }
    }

    /** timestamp when source file was last modified. */
    private Date lastModified = null;

    /** Namespace URI of registry document */
    private static final String REG_XML_NS = "http://www.dspace.org/xmlns/dspace/formatRegistry";

    /** XML namespace for built-in format registries */
    private static final Namespace REG_NS = Namespace.getNamespace("reg", REG_XML_NS);

    /** configuration key for default support level, value is string  */
    private final static String CONFIG_SUPPORTLEVEL_SUFFIX = ".defaultSupportLevel";

    /** lookup the format entries; read once from XML init file. */
    private Map<String,Entry> registry = null;

    /** for identification by filename extension: map of extension to the
     ** list of identifiers of formats it indicates.
     **/
    private Map<String,List<String>> extensionMap = null;

    /** for identification by complete filename: map of filename to the
     ** list of identifiers of formats it indicates.
     **/
    private Map<String,List<String>> filenameMap = null;

    /** cached default support level derived from configuration */
    private int defaultSupportLevel = -1;

    /** log4j logger */
    private static Logger log = Logger.getLogger(BuiltinFormatRegistry.class);

    /**
     * @return  the DSpace format-identifier namespace of this registry.
     */
    abstract public String getNamespace();

    /**
     * Return the URL needed to configure the underlying registry service;
     * this allows the registry to configure itself from the DSpace
     * configuration.
     * For this implementation it is the path to the registry document.
     * @return  URL for the main contact point of the registry resource
     */
    public URL getContactURL()
    {
        String key = "formatRegistry."+getNamespace()+".document";
        String fn = ConfigurationManager.getProperty(key);
        if (fn == null)
        {
            log.warn("format registry not found in configuration, key="+key);
            return null;
        }
        try
        {
            return (fn.startsWith("file:")) ?  new URL(fn) : new URL("file", null, fn);
        }
        catch (MalformedURLException e)
        {
            log.error("Failed to create URL: init filename="+fn);
            return null;
        }
    }

    // read in the registry, at startup or whenever it has changed.
    private void init()
        throws FormatRegistryException
    {
        if (registry == null)
            decacheRegistry();
        else if (lastModified != null)
        {
            long lmi = getRegFile().lastModified();
            if (lmi > 0)
            {
                if (new Date(lmi).after(lastModified))
                    decacheRegistry();
            }
        }
    }

    // returns the path to registry source file -- if none is configured,
    // construct default based on ${dspace.dir}
    private File getRegFile()
        throws FormatRegistryException
    {
        URL ci = getContactURL();
        if (ci == null)
        {
            String ddir = ConfigurationManager.getProperty("dspace.dir");
            if (ddir == null)
                throw new FormatRegistryException ("No initialization file configured for builtin format registry name="+getNamespace());
            return new File(new File(ddir+File.separator+"config"+File.separator+"formats"+File.separator),
                            getNamespace().toLowerCase()+"-formats.xml");
        }
        else
            return new File(ci.getPath());
    }

    /**
     * read and parse the registry, store result in the instance variables
     * registry, extensionMap,filenameMap .
     *
     * @throw FormatRegistryException on any fatal error.
     */
    private void decacheRegistry()
        throws FormatRegistryException
    {
        File regFile = getRegFile();
        long lmi = regFile.lastModified();
        if (lmi > 0)
            lastModified = new Date(lmi);
        log.info("Reading built-in "+getNamespace()+" format registry from source="+regFile);

        String keyPrefix = "formatRegistry."+getNamespace()+".";
        boolean validate = ConfigurationManager.getBooleanProperty(keyPrefix+"validate");
        SAXBuilder builder = new SAXBuilder(validate);

        // Set validation feature
        if (validate)
        {
            builder.setFeature("http://apache.org/xml/features/validation/schema",
                    true);
            String schema = ConfigurationManager.getProperty(keyPrefix+"schema");

            // Tell the parser where local copies of schemas are, to speed up
            // validation.  Local XSDs are identified in the configuration file.
            if (schema != null)
               builder.setProperty(
                       "http://apache.org/xml/properties/schema/external-schemaLocation",
                       REG_XML_NS+" "+schema);
        }

        try
        {
            Document regDoc = builder.build(new FileInputStream(regFile));
            Element regRoot = regDoc.getRootElement();
            List<Element> eentries = (List<Element>)regRoot.getChildren("entry", REG_NS);
            Map<String,Entry> newRegistry = new HashMap<String,Entry>();
            Map<String,List<String>> newExtensionMap = new HashMap<String,List<String>>();
            Map<String,List<String>> newFilenameMap = new HashMap<String,List<String>>();

            for (Element eent : eentries)
            {
                Entry ne = new Entry();
                String name = eent.getChildTextNormalize("name", REG_NS);
                if (name == null)
                    throw new FormatRegistryException("Registry Entry is missing <name> in init file for builtin Format Registry, name="+getNamespace());
                ne.name = name;
                ne.description = eent.getChildTextNormalize("description", REG_NS);
                ne.mimeType = eent.getChildTextNormalize("mime-type", REG_NS);
                Element extsigs = eent.getChild("external-signatures", REG_NS);
                if (extsigs != null)
                {
                    List<String> exts = new ArrayList<String>();
                    List<String> names = new ArrayList<String>();
                    for (Element eext : (List<Element>)extsigs.getChildren("extension", REG_NS))
                    {
                        String extVal = eext.getTextNormalize();
                        exts.add(extVal);
                        String key = extVal.toLowerCase();
                        if (newExtensionMap.containsKey(key))
                            newExtensionMap.get(key).add(name);
                        else
                        {
                            List<String> nl = new ArrayList<String>();
                            nl.add(name);
                            newExtensionMap.put(key, nl);
                        }
                    }
                    ne.extensions = exts.toArray(new String[exts.size()]);

                    // just put "filename" external sigs in map for Identifier
                    for (Element efn : (List<Element>)extsigs.getChildren("filename", REG_NS))
                    {
                        String extVal = efn.getTextNormalize();
                        String key = extVal.toLowerCase();
                        if (newFilenameMap.containsKey(key))
                            newFilenameMap.get(key).add(name);
                        else
                        {
                            List<String> nl = new ArrayList<String>();
                            nl.add(name);
                            newFilenameMap.put(key, nl);
                        }
                    }
                }
                else
                    ne.extensions = new String[0];

                Element extids = eent.getChild("external-identifiers", REG_NS);
                if (extids != null)
                {
                    List<Element> confIds = (List<Element>)extids.getChildren("identifier", REG_NS);
                    List<String> ids = new ArrayList<String>();
                    for (Element eid : confIds)
                    {
                        ids.add(eid.getTextNormalize());
                    }
                    ne.identifiers = ids.toArray(new String[ids.size()]);
                }
                // compensate for missing identifiers in source file
                else
                {
                    ne.identifiers = new String[1];
                    ne.identifiers[0] = FormatRegistryManager.makeIdentifier(getNamespace(), ne.name);
                }
                if (log.isDebugEnabled())
                    log.debug("Adding entry ns="+getNamespace()+", entry="+ne);
                newRegistry.put(ne.name, ne);
            }
            registry = newRegistry;
            extensionMap = newExtensionMap;
            filenameMap = newFilenameMap;
        }
        catch (JDOMException je)
        {
            // failure in initialization is fatal.
            throw new IllegalStateException("Error validating or parsing registry initialization file=\""+regFile+"\", for name="+
                    getNamespace(),  je);
        }
        catch (IOException je)
        {
            // failure in initialization is fatal.
            throw new IllegalStateException("Error Reading registry initialization file for name="+
                    getNamespace(),  je);
        }
    }

    /**
     * Returns synonym identifiers, if any.
     *
     * @param context the database context
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return array of all synonyms for this format in the registry.
     */
    public String[] getSynonymIdentifiers(Context context, String identifier)
        throws FormatRegistryException, AuthorizeException
    {
        init();
        Entry e = registry.get(identifier);
        if (e == null)
            throw new FormatRegistryNotFoundException("Cannot find \""+identifier+"\" in registry "+getNamespace());
        return e.identifiers;
    }

    /**
     * Import a data format, creating a new BitstreamFormat.
     *
     * @param context the database context
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return the new BSF
     * @throws FormatRegistryException for any failure contacting or accessing the registry.
     */
    public BitstreamFormat importExternalFormat(Context context, String id)
        throws FormatRegistryException, AuthorizeException, SQLException
    {
        init();
        Entry e = registry.get(id);
        if (e == null)
            throw new FormatRegistryNotFoundException("Format not found for import in registry name="+getNamespace()+", identifier="+id);

        BitstreamFormat bsf = FormatRegistryManager.createBitstreamFormat(context);
        bsf.setName(e.name);
        if (e.description != null)
            bsf.setDescription(e.description);
        if (e.extensions.length > 0)
            bsf.setCanonicalExtension(e.extensions[0]);
        bsf.setMIMEType((e.mimeType != null) ? e.mimeType : FormatRegistryManager.DEFAULT_MIME_TYPE);
        bsf.addIdentifier(FormatRegistryManager.makeIdentifier(getNamespace(), e.name));
        for (String nsid : e.identifiers)
            bsf.addIdentifier(nsid);
        bsf.setLastImported(getNamespace(), e.name, new Date());
         
        bsf.setSupportLevel(getDefaultSupportLevel());
         
        return bsf;
    }

    /**
     * Set BSF's members to values of registry entry, unless they have
     * the override bit set.
     * <p>
     * @param context the database context
     * @param bsf the BitstreamFormat to update
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @throws FormatRegistryException for any failure contacting or accessing the registry, or if the identifier cannot be found.
     */
    public void updateBitstreamFormat(Context context, BitstreamFormat bsf, String id)
        throws FormatRegistryException, AuthorizeException, SQLException
    {
        init();

        Entry e = registry.get(id);
        if (e == null)
            throw new FormatRegistryNotFoundException("No such External Format, in registry name="+getNamespace()+", identifier="+id);

        // check for any "new" identifier collisions
        for (String nsid : e.identifiers)
        {
            BitstreamFormat other = BitstreamFormat.findByIdentifier(context,
                                                                     FormatRegistryManager.namespaceOf(nsid),
                                                                     FormatRegistryManager.identifierOf(nsid), false);
            if (other != null && !other.equals(bsf))
                throw new FormatRegistryException("Cannot update "+bsf+
                  " from builtin "+getNamespace()+" registry, since the identifier "+nsid+" it references is already bound to a different format: "+other);
        }

        if (!bsf.isOverrideName())
            bsf.setName(e.name);
        if (e.description != null && !bsf.isOverrideDescription())
            bsf.setDescription(e.description);
        if (e.mimeType != null && !bsf.isOverrideMIMEType())
            bsf.setMIMEType(e.mimeType);
     
        // look for identifiers that aren't in the BSF already.
        String bids[] = bsf.getIdentifiers();
    identifierInBSF:
        for (String nsid : e.identifiers)
        {
            for (String bid : bids)
                if (bid.equals(nsid))
                    continue identifierInBSF;
     
            bsf.addIdentifier(nsid);
        }
     
        // check canonical file extension, make it the same.
        if (!bsf.isOverrideCanonicalExtension())
        {
            if (e.extensions.length > 0)
                bsf.setCanonicalExtension(e.extensions[0]);
            else
                bsf.setCanonicalExtension(null);
        }
     
        bsf.setLastImported(getNamespace(), e.name, new Date());
    }

    /**
     * Returns last-modified date of the initialization file, for all entries.
     *
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return date/time when this entry was last changed, or null if not available.
     * @throws FormatRegistryException upon failure accessing the registry.
     */
    public Date getLastModified(String identifier)
        throws FormatRegistryException
    {
        init();
        if (registry.get(identifier) == null)
            throw new FormatRegistryNotFoundException("No such External Format, in registry name="+getNamespace()+", identifier="+identifier);
        return lastModified;
    }

    /**
     * Predicate, true if format named by sub is a subtype or
     * otherwise "conforms" to the format defined by fmt.
     *
     * Since the builtin registry data model has no
     * notion of relationships between formats, the only way we
     * know if one format "conforms to" another is if they are identical.
     *
     * @param sub bare (NON-namespaced) external identifier of the subtype format
     * @param fmt bare (NON-namespaced) external identifier of the supertype format
     * @throws FormatRegistryException upon failure accessing the registry.
     */
    public boolean conformsTo(String sub, String fmt)
        throws FormatRegistryException
    {
        init();
        if (!registry.containsKey(sub))
            throw new FormatRegistryNotFoundException("No such External Format, in registry name="+getNamespace()+", identifier="+sub);
        if (!registry.containsKey(fmt))
            throw new FormatRegistryNotFoundException("No such External Format, in registry name="+getNamespace()+", identifier="+fmt);

        return sub.equalsIgnoreCase(fmt);
    }

    /**
     * Free any resources associated with this registry instance.
     */
    public void shutdown()
    {
    }


    /**
     * Get the DEFAULT support level applied to a new bitstream format
     * created by importExternalFormat().  Will be one of
     * <code>UNKNOWN</code>,<code>KNOWN</code> or <code>SUPPORTED</code>.
     * <p>
     * NOTE: This should eventually be decided by a policy engine, rather
     * than a hardwired default in the configuration.
     * <p>
     * @return the default support level
     */
    private int getDefaultSupportLevel()
    {
        if (defaultSupportLevel < 0)
        {
            String key = "formatRegistry."+getNamespace()+CONFIG_SUPPORTLEVEL_SUFFIX;
            String name = ConfigurationManager.getProperty(key);
            if (name == null)
                defaultSupportLevel = BitstreamFormat.UNKNOWN;
            else
            {
                name = name.trim();
                for (int i = 0; i < BitstreamFormat.supportLevelText.length; ++i)
                {
                    if (name.equalsIgnoreCase(BitstreamFormat.supportLevelText[i]))
                    {
                        defaultSupportLevel = i;
                        break;
                    }
                }
                if (defaultSupportLevel < 0)
                {
                    log.error("BitstreamFormat support level configuration value unrecognized: "+key+"=\""+name+"\"");
                    defaultSupportLevel = BitstreamFormat.UNKNOWN;
                }
            }
        }
        return defaultSupportLevel;
    }

    /**
     * Implement format identification plugin as well.
     * Try matching both the extension and the entire filename against
     * the maps set up with the configuration was read.
     * Matches are NOT case-sensitive, because many server filesystems are not.
     * <p>
     * @param context the DB context
     * @param target the Bitstream to identify
     * @param results List of results produced by previous identifyFormat methods
     * @return list of results, possibly modified or replacing the result param
     * @throws FormatIdentifierException when a fatal error in the process
     *   of identifying the format prevents it from returning conclusive results
     */
    public List<FormatHit> identifyFormat(Context context, Bitstream target, List<FormatHit> results)
        throws FormatIdentifierException, AuthorizeException
    {
        try
        {
            init();
        }
        catch (FormatRegistryException e)
        {
            // failure to initialize should be fatal.
            throw new IllegalStateException(e);
        }
        String filename = target.getName();
        if (filename != null)
        {
            int lastDot = filename.lastIndexOf('.');
            if (lastDot != -1)
            {
                String extension = filename.substring(lastDot + 1);
                if (extension.length() > 0)
                    addHits(extensionMap.get(extension.toLowerCase()), results);
            }
            int lastSlash = filename.lastIndexOf('/');
            String fname = (lastSlash < 0) ? filename : filename.substring(lastSlash+1);
            addHits(filenameMap.get(fname.toLowerCase()), results);
        }
        return results;
    }

    // add hits for each listed identified format; OK if list is null.
    private void addHits(List<String> idents, List<FormatHit> results)
    {
        if (idents != null)
        {
            String myname = this.getClass().getName();
            for (String ident : idents)
            {
                FormatHit hit = new FormatHit(getNamespace(),
                                              ident,
                                              ident,
                                              FormatConfidence.CIRCUMSTANTIAL,
                                              myname,
                                              null);
                hit.addToResults(results);
            }
        }
    }

    /**
     * Get list of available identifiers; this is an extra function
     * included to help the conversion process from old DSpace DBs.
     */
    public Set<String> getAllIdentifiers()
        throws FormatRegistryException
    {
        init();
        return registry.keySet();
    }
}
