/*
 * CSSIdentifier.java
 *
 * Version: $Revision: 1523 $
 *
 * Date: $Date: 2006-05-26 10:23:17 -0400 (Fri, 26 May 2006) $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.content.format;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.util.List;
import java.util.ListIterator;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;

import org.apache.log4j.Logger;

import org.w3c.css.sac.Parser;
import org.w3c.css.sac.InputSource;
import org.w3c.css.sac.CSSException;
import org.w3c.css.sac.CSSParseException;
import org.w3c.css.sac.DocumentHandler;
import org.w3c.css.sac.ErrorHandler;
import org.w3c.css.sac.SACMediaList;
import org.w3c.css.sac.SelectorList;
import org.w3c.css.sac.LexicalUnit;

import org.dspace.authorize.AuthorizeException;
import org.dspace.core.ConfigurationManager;
import org.dspace.core.Context;
import org.dspace.content.Bitstream;

/**
 * This identifier solves a tricky problem with identifying CSS bitstreams
 * correctly:  Since the format is too loosely specified to support an
 * internal-signature method, it gets identified _heuristically_ as plain
 * text, and that takes precedence over a "circumstantial" ID by the
 * filename extension.
 *
 * At the same time, it is essential to identify HTML Cascading
 * Stylesheet bitstreams accurately so they are given the the correct
 * MIME content-type on Web dissemination, otherwise archived websites
 * will not be displayed correctly.
 *
 * This method attempts to _parse_ a bitstream as CSS and upon success, gives
 * it a hit with a higher confidence (positive-specific) value.  It only
 * attempts to refine the identification when there is already a CSS
 * hit present in the results, so run the method that produces circumstantial
 * hits first (e.g. DROID).
 *
 * It uses an SAC parser implementation, see
 *  http://www.w3.org/Style/CSS/SAC/
 *
 * This has been tested with flute 1.3, available on:
 *  http://www.w3.org/Style/CSS/SAC/
 *  http://www.w3.org/2002/06/flutejava-1.3.zip
 *
 * Configuration:
 *   formatIdentifier.CSSIdentifier.parser = classname
 *    Must be the class implementing a SAC CSS parser, org.w3c.css.sac.Parser
 *    E.g.: formatIdentifier.CSSIdentifier.parser = org.w3c.flute.parser.Parser
 *
 *   formatIdentifier.CSSIdentifier.identifiers = NSID, ...
 *    value is a comma-separated list of namespaced external format identifiers
 *    which represent the Cascading Stylesheet (CSS) format.  For example,
 *    value could be: PRONOM:x-fmt/224, DSpace:CSS
 *    NOTE: the first value is taken to be the canonical format which is
 *    assigned to any hit that gets generated.
 *    This list should include all values for CSS which are likely to be
 *    returned by methods in your stack.
 *
 * @author Larry Stone
 * @version $Revision: 1523 $
 * @see FormatIdentifier
 */
public class CSSIdentifier
    implements FormatIdentifier
{
    private static Logger log = Logger.getLogger(CSSIdentifier.class);

    // class to instantiate for SAC parser, get from config
    private static Class parserClass = null;

    // configuration key for parser implementation
    private final static String CSS_PARSER_KEY =
            "formatIdentifier.CSSIdentifier.parser";

    // acceptable external identifiers that indicate CSS
    private static Set cssIdentifiers = null;

    // the canonical external identifier to apply when adding a hit for CSS.
    private static String canonicalCSS = null;

    // configuration key to lookup list of CSS external format identifiers.
    private final static String CSS_IDS_KEY =
            "formatIdentifier.CSSIdentifier.identifiers";

    // initialize from config; returns true if it fails
    private static boolean init()
    {
        boolean result = false;
        if (cssIdentifiers == null)
        {
            String rawIDs = ConfigurationManager.getProperty(CSS_IDS_KEY);
            if (rawIDs == null)
            {
                log.error("No configuration value for key="+CSS_IDS_KEY+", so CSSIdentifier is inactive.");
                result = true;
            }
            else
            {
                cssIdentifiers = new HashSet();
                String nsids[] = rawIDs.trim().split("\\s*,\\s*");
                canonicalCSS = nsids[0];
                for (String id : nsids)
                    cssIdentifiers.add(id);
            }
        }
        if (parserClass == null)
        {
            String name = ConfigurationManager.getProperty(CSS_PARSER_KEY);
            if (name == null)
            {
                log.error("No configuration value for key="+CSS_PARSER_KEY+", so CSSIdentifier is inactive.");
                result = true;
            }
            else
            {
                try
                {
                    parserClass = Class.forName(name);
                }
                catch (ClassNotFoundException e)
                {
                    log.error("Failed to lookup class (name="+name+") for CSS parser:",e);
                    result = true;
                }
            }
        }
        return result;
    }

    /**
     * First check if Bitstream was already identified as a CSS format
     * at a lower level of confidence, e.g. CIRCUMSTANTIAL.
     * Then attempt to parse the Bitstream as CSS, add a hit at the
     * VALIDATED confidence level if it succeeds.
     * This is really more of a "validator" than "identifier" method.
     * <p>
     * @param context the DB context
     * @param target the Bitstream to identify
     * @param results List of results produced by previous identifyFormat methods
     * @return list of results, possibly modified or replacing the result param
     * @throws FormatIdentifierException when a fatal error in the process
     *   of identifying the format prevents it from returning conclusive results
     */
    public List<FormatHit> identifyFormat(Context context, Bitstream bitstream, List<FormatHit> results)
        throws FormatIdentifierException, AuthorizeException, SQLException, IOException
    {
        if (init())
            return results;

        // look for circumstantial CSS entry, if found, try to upgrade
        // by attempting to parse a CSS document.
        for (FormatHit fh : results)
        {
            if (fh.confidence == FormatConfidence.CIRCUMSTANTIAL &&
                cssIdentifiers.contains(fh.getIdentifier()))
            {
                if (parseCSS(bitstream.retrieve()))
                {
                    new FormatHit(FormatRegistryManager.namespaceOf(canonicalCSS),
                                  FormatRegistryManager.identifierOf(canonicalCSS),
                                  "Cascading Style Sheet",
                                  FormatConfidence.VALIDATED,
                                  this.getClass().getName(),
                                  null).addToResults(results);
                    return results;
                }
            }
        }
        return results;
    }

    // returns true only if input was successfully parsed as CSS.
    private boolean parseCSS(InputStream css)
        throws IOException
    {
        try
        {
            // Parser parser = new SACParser();
            Parser parser = (org.w3c.css.sac.Parser)parserClass.newInstance();
            validateHandler dh = new validateHandler();
            parser.setDocumentHandler(dh);

            // although there *should* be a default error handler that
            // throws CSSParseException, with flute-1.3 we must implement it:
            parser.setErrorHandler(new simpleErrorHandler());
            parser.parseStyleSheet(new InputSource(new InputStreamReader(css)));
            return dh.ended;
        }
        catch (InstantiationException e)
        {
            log.error(e);
            return false;
        }
        catch (IllegalAccessException e)
        {
            log.error(e);
            return false;
        }
        catch (CSSException e)
        {
            if (log.isDebugEnabled())
                log.debug("Failed parsing CSS, assuming this is NOT a stylesheet: "+ e);
            return false;
        }
    }

    // SAC handler that does nothing; some handler is required by parser.
    public class validateHandler implements DocumentHandler
    {
        public boolean ended = false;

        public void startDocument(InputSource source)
            throws CSSException
        {
        }
        public void endDocument(InputSource source) throws CSSException
        {
            ended = true;
        }
     
        public void comment(String text) throws CSSException
        {
        }
     
        public void ignorableAtRule(String atRule) throws CSSException
        {
        }
     
        public void namespaceDeclaration(String prefix, String uri)
            throws CSSException
        {
        }
     
        public void importStyle(String uri, SACMediaList media,
                                String defaultNamespaceURI)
            throws CSSException
        {
        }
     
        public void startMedia(SACMediaList media) throws CSSException
        {
        }
     
        public void endMedia(SACMediaList media) throws CSSException
        {
        }
     
        public void startPage(String name, String pseudo_page) throws CSSException
        {
        }
     
        public void endPage(String name, String pseudo_page) throws CSSException
        {
        }
     
        public void startFontFace() throws CSSException
        {
        }
     
        public void endFontFace() throws CSSException
        {
        }
     
        public void startSelector(SelectorList selectors) throws CSSException
        {
        }
     
        public void endSelector(SelectorList selectors) throws CSSException
        {
        }
     
        public void property(String name, LexicalUnit value, boolean important)
            throws CSSException
        {
        }
    }

    // flute 1.3 won't throw errors w/o explicit error handler
    public class simpleErrorHandler implements ErrorHandler
    {
        public void warning(CSSParseException exception)
            throws CSSException
        {
            if (log.isDebugEnabled())
                log.debug("Got CSS parse warning: "+exception);
        }
        public void error(CSSParseException exception)
            throws CSSException
        {
            throw exception;
        }
        public void fatalError(CSSParseException exception)
            throws CSSException
        {
            throw exception;
        }
    }
}
