/*
 * DROIDIdentifier.java
 *
 * Version: $Revision: 1523 $
 *
 * Date: $Date: 2006-05-26 10:23:17 -0400 (Fri, 26 May 2006) $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.content.format;

import java.io.File;
import java.io.IOException;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import uk.gov.nationalarchives.droid.AnalysisController;
import uk.gov.nationalarchives.droid.signatureFile.FFSignatureFile;
import uk.gov.nationalarchives.droid.IdentificationFile;
import uk.gov.nationalarchives.droid.binFileReader.AbstractByteReader;
import uk.gov.nationalarchives.droid.binFileReader.ByteReader;
import uk.gov.nationalarchives.droid.binFileReader.FileByteReader;
import uk.gov.nationalarchives.droid.signatureFile.FileFormat;
import uk.gov.nationalarchives.droid.FileFormatHit;

import org.apache.log4j.Logger;

import org.dspace.authorize.AuthorizeException;
import org.dspace.content.Bitstream;
import org.dspace.core.ConfigurationManager;
import org.dspace.core.Context;

/**
 * Identify bitstream formats using DROID (see http://droid.sourceforge.net).
 *
 * Configuration:
 *   formatIdentifier.DROID.signatureFile = <path to sig file>
 *
 * @author Larry Stone
 * @version $Revision: 1523 $
 * @see FormatIdentifier
 */
public class DROIDIdentifier
    implements FormatIdentifier
{
    private static Logger log = Logger.getLogger(DROIDIdentifier.class);

    // configuration file key for signature file
    private static final String SIGNATURE_FILE_KEY = "formatIdentifier.DROID.signatureFile";

    // DROID object for a loaded signature file
    private FFSignatureFile droidSigFile = null;

    // translate DROID's hit type to an ordered set of integers to allow
    // fast comparisons and sorting.
    private static int hitType2order[];

    // translate DROID hit type to DSpace FormatConfidence
    private static FormatConfidence hitType2confidence[];

    // Initialize tables that allow sorting of hit types (hitType2order)
    // and conversion of DROID hits to DSpace confidence (hitType2confidence).
    // Since these are all low integers it is a simple matter to use an array
    // to convert them rather than a Map, and much more efficient too.
    static
    {
        int nTypes = Math.max(AnalysisController.HIT_TYPE_POSITIVE_SPECIFIC,
                       Math.max(AnalysisController.HIT_TYPE_POSITIVE_GENERIC,
                         Math.max(AnalysisController.HIT_TYPE_POSITIVE_GENERIC_OR_SPECIFIC,
                                  AnalysisController.HIT_TYPE_TENTATIVE)));
        hitType2order = new int[nTypes+1];
        Arrays.fill(hitType2order, -1);
        hitType2order[AnalysisController.HIT_TYPE_POSITIVE_SPECIFIC] = 1;
        hitType2order[AnalysisController.HIT_TYPE_POSITIVE_GENERIC] = 2;
        hitType2order[AnalysisController.HIT_TYPE_POSITIVE_GENERIC_OR_SPECIFIC] = 3;
        hitType2order[AnalysisController.HIT_TYPE_TENTATIVE] = 4;

        // fill in confidence translation table
        hitType2confidence = new FormatConfidence[nTypes+1];
        Arrays.fill(hitType2confidence, FormatConfidence.UNIDENTIFIED);
        hitType2confidence[AnalysisController.HIT_TYPE_POSITIVE_SPECIFIC] =
            FormatConfidence.POSITIVE_SPECIFIC;
        hitType2confidence[AnalysisController.HIT_TYPE_POSITIVE_GENERIC] =
            FormatConfidence.POSITIVE_GENERIC;
        hitType2confidence[AnalysisController.HIT_TYPE_POSITIVE_GENERIC_OR_SPECIFIC] =
            FormatConfidence.POSITIVE_GENERIC;
        hitType2confidence[AnalysisController.HIT_TYPE_TENTATIVE] =
            FormatConfidence.CIRCUMSTANTIAL;
    }

    // returns cached DROID signature file
    private FFSignatureFile getSigFile()
        throws FormatIdentifierException, IOException
    {
        if (droidSigFile == null)
        {
            try
            {
                AnalysisController ac = new AnalysisController();
                String path = ConfigurationManager.getProperty(SIGNATURE_FILE_KEY);
                if (path == null)
                    throw new IOException("No DROID signature file path, check configuration value for: "+SIGNATURE_FILE_KEY);
                log.debug("Attempting to read DROID file format signature file: "+path);
                droidSigFile = ac.parseSigFile(path);
                droidSigFile.prepareForUse();
                log.info("Loaded DROID signature file version \""+
                         droidSigFile.getVersion()+"\", from path="+path);
            }
            catch (Exception e)
            {
                log.error("Failed getting DROID signature file:",e);
                IOException ie = new IOException();
                ie.initCause(e);
                throw ie;
            }
        }
        return droidSigFile;
    }

    /**
     * Call DROID to identify the format of a Bitstream.
     * Since DROID 2.0 has no way to process a random InputStream (it only
     * allows the standard input, System.in), we created the custom ByteReader
     * DROIDAnyStreamByteReader to do that.
     * XXX TODO AS FUTURE IMPROVEMENT:  it would improve performance if
     * we give DROID access to the File containing the bitstream so it
     * can be opened for random access.  Unfortunately this would break the
     * modularity of BitstreamStorageManager.
     * <p>
     * @param context the DB context
     * @param target the Bitstream to identify
     * @param results List of results produced by previous identifyFormat methods
     * @return list of results, possibly modified or replacing the result param
     * @throws FormatIdentifierException when a fatal error in the process
     *   of identifying the format prevents it from returning conclusive results
     */
    public List<FormatHit> identifyFormat(Context context, Bitstream bitstream, List<FormatHit> results)
        throws FormatIdentifierException, AuthorizeException, SQLException, IOException
    {
        FFSignatureFile mySigFile = getSigFile();
        String path = bitstream.getName();
        if (path == null)
            path = "";
         
        IdentificationFile idf = new IdentificationFile(path);

        ByteReader testFile = null;
        DROIDAnyStreamByteReader sr = new DROIDAnyStreamByteReader(idf, bitstream.retrieve());
        File tempPath = sr.getTempFile();
        if (tempPath == null) {
            testFile = sr;
        } else {
            // KLUDGE:
            // In order to get a file-oriented ByteReader, we
            // temporarily set the IDFile's path to the temp file's path,
            // to force newByteReader() to make the right kind of reader.
            // Then change it back after creating the ByteReader so the
            // Bitstream's correct file extension is seen by DROID.
            // This wouldn't be so difficult if the ByteReader factory wasn't
            // so inexplicably bizarre..
            String oldIdPath = idf.getFilePath();
            idf.setFilePath(tempPath.getPath());
            testFile = AbstractByteReader.newByteReader(idf, true);
            idf.setFilePath(oldIdPath);
        }

        try
        {
            mySigFile.runFileIdentification(testFile);
        }
        catch (Exception e)
        {
            log.error("DROID runFileIdentification() got exception on Bitstream ID="+bitstream.getID(), e);
            return results;
        }
        finally
        {
            // XXX KLUDGE: clean up after DROID, there is a BUG that it doesn't!
            if (tempPath != null)
                tempPath.delete();
        }

        // put DROID hits into an Array so we can sort them
        int nHits = idf.getNumHits();
        if (log.isDebugEnabled())
            log.debug("Got "+String.valueOf(nHits)+" format hits from DROID.");
        FileFormatHit hits[] = new FileFormatHit[nHits];
        for (int i = 0; i < nHits; ++i)
            hits[i] = idf.getHit(i);
        Arrays.sort(hits, new CompareHit());
        String myname = this.getClass().getName();
        
        // translate to DSpace FormatHits and add to results, in order:
        for (FileFormatHit hit : hits)
        {
            // normalize warning to null if it's empty string.
            String wng = hit.getHitWarning();
            if (wng != null && wng.length() == 0)
                wng = null;
            FileFormat hff = hit.getFileFormat();
            new FormatHit(PRONOMFormatRegistry.NAMESPACE,
                          hff.getPUID(),
                          hff.getName(),
                          hitType2confidence[hit.getHitType()],
                          myname,
                          wng).addToResults(results);
        }
        return results;
    }

    // comparator for DROID's FileFormatHit, to sort its results by our criteria.
    static class CompareHit
        implements Comparator<FileFormatHit>
    {
        public CompareHit()
        {
            super();
        }

        /**
         * Compare along axes:
         *  - status: postive, tenative, negative
         *  - specificiity: specific, generic (only for positive)
         *  - whether one format has priority over the other.
         *  - warnings: none, any
         */
        public int compare(FileFormatHit a, FileFormatHit b)
        {
            int aType = hitType2order[a.getHitType()];
            int bType = hitType2order[b.getHitType()];
            if (aType < bType)
                return -1;
            else if (aType > bType)
                return 1;

            // same type of hit.. look for priorities, or warnings.
            else
            {
                FileFormat aFmt = a.getFileFormat();
                FileFormat bFmt = b.getFileFormat();
                if (hasPriorityOver(aFmt, bFmt))
                    return -1;
                if (hasPriorityOver(bFmt, aFmt))
                    return 1;
                if (a.getHitWarning() == null)
                {
                    // a has no warning, b has warning.
                    if (b.getHitWarning() != null)
                        return -1;
                }
                else
                {
                    // a has warning, b has none
                    if (b.getHitWarning() == null)
                        return 1;
                }
                return 0;
            }
        }

        // return true if a has priority over b.
        private boolean hasPriorityOver(FileFormat a, FileFormat b)
        {
            int nPrio = a.getNumHasPriorityOver();
            int bID = b.getID();
            for (int i = 0; i < nPrio; ++i)
            {
                if (a.getHasPriorityOver(i) == bID)
                {
                    if (log.isDebugEnabled())
                        log.debug("Found higher priority format: PUID="+a.getPUID()+" > PUID="+b.getPUID());
                    return true;
                }
            }
            return false;
        }

        // doesn't make sense
        public boolean equals(Object other)
        {
            return false;
        }
    }
}
