package org.dspace.content.format;

import org.dspace.content.BitstreamFormat;
import org.dspace.content.Bitstream;
import org.dspace.core.Context;
import org.dspace.authorize.AuthorizeException;

import java.sql.SQLException;
import java.util.List;
import java.util.ArrayList;
import org.apache.log4j.Logger;

/**
 * FormatHit is a data structure to contain the results of a single
 * format identification by a FormatIdentifier method; "hit" means a
 * match against identification criteria of some sort.
 * It collects data required to compare the different hits to sort them
 * appropriately in the list of results, and to apply the selected hit to
 * the Bitstream it identifies.
 * <p>
 * @author Larry Stone
 * @version $Revision: 1323 $
 * @see FormatIdentifier
 */
public class FormatHit implements Comparable<FormatHit>
{
    /** log4j logger */
    private static Logger log = Logger.getLogger(FormatHit.class);

    /**
     * Namespace and identifier of the external format
     * (which may not yet exist as a BitstreamFormat)
     */
    public String namespace = null;
    public String identifier = null;

    /**
     * Human-readable name of the format to aid in selection.
     */
    public String name = null;
    
    /**
     * Measure of the "confidence" of this hit, one of the constants
     *  described below.
     */
    public FormatConfidence confidence;
    
    /**
     * Name (e.g. class) of the originator of the identified format.
     */
    public String source;

    /**
     * Any warning from the identification process (to show to user)
     */
    public String warning = null;

    /**
     * Constructor
     */
    public FormatHit(String namespace, String identifier, String name, FormatConfidence confidence, String source, String warning)
    {
        this.namespace = namespace;
        this.identifier = identifier;
        this.name = name;
        this.confidence = confidence;
        this.warning = warning;
        this.source = source;
    }

    /**
     * Utility to add this hit in the correct place on a results list.
     *
     * NOTE that this is only the DEFAULT algorithm; an identifier
     * method can implement whatever algorithm it chooses, e.g. cleaning
     * out all lower-valued hits, replacing the list, whatever..
     *
     * - if list is empty or this is best hit, put at head of list.
     * - if this is equal to best hit, put second.
     * - otherwise, append to end of list.
     *
     * @param results list of result hits.
     */
    public void addToResults(List<FormatHit> results)
    {
        if (results.size() == 0)
        {
            results.add(this);
        }
        else
        {
            int comp = this.compareTo(results.get(0));
            if (comp > 0)
                results.add(0, this);
            else if (comp == 0)
                results.add(1, this);
            else
                results.add(this);
        }
    }

    /**
     * Convenience routine to return a single string containing the namespaced format identifier:
     * @return the namespaced identifier
     */
    public String getIdentifier()
    {
        return FormatRegistryManager.makeIdentifier(namespace, identifier);
    }

    /**
     * Convenience routine to return the BitstreamFormat object.
     * @param context the DB context
     * @return BSF of this hit or NULL if it cannot be found/imported.
     */
    public BitstreamFormat getBitstreamFormat(Context context)
        throws SQLException, FormatRegistryException, AuthorizeException
    {
        return BitstreamFormat.findByIdentifier(context, namespace, identifier, true);
    }

    /**
     * Comparison method to satisfy Comparable<FormatHit>.
     */
    public int compareTo(FormatHit other)
    {
        FormatRegistry thisReg = FormatRegistryManager.find(this.namespace);
        FormatRegistry otherReg = FormatRegistryManager.find(other.namespace);

        // if both namespaces are resolved or both not, compare confidence..
        if ((thisReg != null && otherReg != null) ||
            (thisReg == null && otherReg == null))
        {
            int result = this.confidence.compareTo(other.confidence);
            if (result == 0)
            {
                // if either has the only warning, that shifts the balance
                if (other.warning != null && this.warning == null)
                    result = 1;
                else if (this.warning != null && other.warning == null)
                    result = -1;
            }
            return result;
        }

        // otherwise give precedence to whichever hit has a resolvable namespace.
        else
            return (thisReg != null) ? 1 : -1;
    }

    /**
     * Set the relevant fields in Bitstream with contents of this hit.
     * @param context the DB context
     * @param bs bitstream to apply to.
     */
    public void applyToBitstream(Context c, Bitstream bs)
        throws SQLException, AuthorizeException, FormatRegistryException
    {
        BitstreamFormat bsf = getBitstreamFormat(c);
        if (bsf == null)
        {
            bs.setFormat(BitstreamFormat.findUnknown(c));
            bs.setFormatConfidence(FormatConfidence.UNIDENTIFIED);
        }
        else
        {
            bs.setFormatConfidence(confidence);
            bs.setFormat(bsf);
        }
        bs.setFormatSource(source);
        bs.update();
    }

    // prettyprint for debugging
    public String toString()
    {
        return "<FormatHit ext_id="+FormatRegistryManager.makeIdentifier(namespace,identifier)+
               ", confidence="+confidence+", warning=\""+(warning==null?"":warning)+"\">";
    }
}
