package org.dspace.content.format;

import java.sql.SQLException;
import java.io.IOException;
import java.util.List;

import org.dspace.core.Context;
import org.dspace.content.Bitstream;
import org.dspace.authorize.AuthorizeException;

/**
 * Interface describing the format identification plugin.
 * <p>
 * @author Larry Stone
 * @version $Revision: 1323 $
 * @see FormatHit, FormatIdentifiermanager
 */
public interface FormatIdentifier
{
    /**
     * Attempt to identify the data format of the given Bitstream and add any
     * conclusions to the <i>results</i> list.  FormatIdentifier is a
     * sequence plugin, the methods are called in order by the
     * FormatIdentifierManager, so one implementation may expect to see
     * results of a previous call and build upon them, e.g. by refining
     * or validating identification results of other methods.
     * <p>
     * An identifier method may use any characteristics of the Bitstream as
     * clues, including the name (and filename extension), contents, size,
     * etc.
     * <p>
     * Note that each implementation is responsible for adding any results
     * it produces to the <i>results</i> list, according to a method or
     * algorithm of its own choosing.  It <i>may</i> simply use the default
     * method in FormatHit.addToResults() but that is only one possible choice.
     * <p>
     * An implementation does not even have to look for formats, it
     * may simply weed out or reorganize previous results.
     * <p>
     * It should throw a FormatIdentifierException when a fatal error
     * or failure accessing resources (e.g. in a filesystem or on the network)
     * prevents it from returning conclusive results.
     * If it simply fails to identify a format, it should do nothing but
     * return the given results list.
     * <p>
     * @param context the DB context
     * @param target the Bitstream to identify
     * @param results List of results produced by previous identifyFormat methods
     * @return list of results, possibly modified or replacing the result param
     * @throws FormatIdentifierException when a fatal error in the process
     *   of identifying the format prevents it from returning conclusive results
     */
    public List<FormatHit> identifyFormat(Context context, Bitstream target, List<FormatHit> results)
        throws FormatIdentifierException, AuthorizeException, SQLException, IOException;
}
