package org.dspace.content.format;

import org.dspace.content.BitstreamFormat;
import org.dspace.content.Bitstream;
import org.dspace.core.ConfigurationManager;
import org.dspace.core.SelfNamedPlugin;
import org.dspace.core.Context;
import org.dspace.core.Utils;
import org.dspace.authorize.AuthorizeException;

import java.net.URL;
import java.net.MalformedURLException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FilterInputStream;
import java.io.IOException;
import java.text.ParseException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.sql.SQLException;
import java.net.URI;
import java.net.URISyntaxException;

import org.jdom.input.SAXBuilder;
import org.jdom.JDOMException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.XMLReader;
import org.xml.sax.InputSource;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.ContentHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.apache.log4j.Logger;

import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.StatusLine;

/**
 * Interface to the PRONOM online registry of technical metadata.
 * PRONOM, see http://www.nationalarchives.gov.uk/pronom/ , is supported
 * by the National Archives of the UK.  As of November 2007, it is the
 * most extensive publically-accessible source of file-format technical
 * metadata, describing over 500 formats.  The DROID format identifying tool
 * (see http://droid.sourceforge.net/ ) is based on PRONOM metadata.
 * <p>
 * This registry is integrated with DSpace as the "PRONOM" namespace.
 * External format identifiers are the PUIDs within PRONOM, e.g. "fmt/3".
 * They are unique and persistent.
 * <p>
 * See comments on the individual methods for more about how PRONOM's
 * sophisticated format data model is translated to the DSpace plugin API.
 * <p>
 * This implementation accesses PRONOM over the Internet, but caches results
 * locally to reduce reliance on the network and improve performance.
 * It downloads two types of files:  First, an "index" that lists all formats
 * and acts as a translation table from the Web-UI "FormatID" to PUID; and
 * second, the PRONOM record for each individual format.  Since the format
 * records must be requested by integer FormatID, the index table must be
 * downloaded first and kept up-to-date.
 * <p>
 * There is also a command-line UI to forcibly decache (update) the local
 * copies of the format index and individual format records, see main().
 * <p>
 * DSpace Configuration keys:<br>
 *
 *  # base URL of PRONOM site, default shown here:
 *  formatRegistry.PRONOM.contact = http://www.nationalarchives.gov.uk/
 *
 *  # default support level of newly imported entries.
 *  formatRegistry.PRONOM.defaultSupportLevel = known
 *
 *  # path to directory where cached files are stored.
 *  formatRegistry.PRONOM.dir = ${dspace.dir}/config/formats/PRONOM
 * <p>
 * @author Larry Stone
 * @version $Revision: 1323 $
 */
public class PRONOMFormatRegistry
    extends SelfNamedPlugin
    implements FormatRegistry
{
    /** DSpace format namespace - by convention, available as public variable */
    public static String NAMESPACE = "PRONOM";

    /** Lookup table of PRONOM PUID to the integer "ID" in the XML dumps.
     ** This appears in the FormatID element of the List & Detail report.
     **/
    private static Map<String,String> puid2formatID = null;

    // configuration key for default support level, value is string
    private final static String CONFIG_SUPPORTLEVEL_SUFFIX = ".defaultSupportLevel";

    // cached default support level derived from configuration
    private int defaultSupportLevel = -1;

    // config key suffix for PRONOM data directory
    private final static String CONFIG_DIRECTORY_SUFFIX = ".dir";

    // path to data directory
    private static File dataDirectory = null;

    // name of file with list of formats..
    private final static String INDEX_FILE = "index.xml";

    // cached contact URL
    private URL contactURL = null;

    // log4j logger
    private static Logger log = Logger.getLogger(PRONOMFormatRegistry.class);

    /**
     * @return  the DSpace format-identifier namespace of this registry.
     */
    public String getNamespace()
    {
        return NAMESPACE;
    }

    /**
     * Return the URL needed to configure the underlying registry service;
     * this allows other users of the registry (e.g. in UIs) to share the
     * same configuration.
     * <p>
     * In the case of PRONOM, the URL is the root of the HTTP URL of the
     * website where the PRONOM UI and the PUID download service reside.
     * It only includes the hostname since it is followed by a "pronom"
     * or "PRONOM" path element depending on which service is chosen.
     * <p>
     * @return  URL for the main contact point of the registry resource
     */
    public URL getContactURL()
    {
        if (contactURL == null)
        {
            String key = "formatRegistry."+getNamespace()+".contact";
            String fn = ConfigurationManager.getProperty(key);
            if (fn == null)
            {
                log.warn("format registry not found in configuration, key="+key);
                return null;
            }
             
            try
            {
                contactURL = (fn.indexOf(":") >= 0) ?  new URL(fn) : new URL("file", null, fn);
            }
            catch (MalformedURLException e)
            {
                log.error("Failed to create URL, value="+fn, e);
            }
        }
        return contactURL;
    }

    /**
     * Gets all external identifiers known to be synonyms of the given
     * identifier, in this registry.
     * The PRONOM implementation reads the file format record for the
     * given PUID, collecting "details" through a special SAX handler.
     * <p>
     * @param context the database context
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return array of all synonyms for this format in the registry.
     */
    public String[] getSynonymIdentifiers(Context context, String identifier)
        throws FormatRegistryException, AuthorizeException
    {
        init();
        FormatDetailHandler dh = getEntryDetails(identifier);
        return dh.identifiers.toArray(new String[dh.identifiers.size()]);
    }


    /**
     * Import a new data format from the registry, creating a new
     * BitstreamFormat.
     *
     * @param context the database context
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return the new BSF, or null if the external format cannot befound.
     * @throws FormatRegistryException for any failure contacting or accessing the registry.
     */
    public BitstreamFormat importExternalFormat(Context context, String puid)
        throws FormatRegistryException, AuthorizeException, SQLException
    {
        init();

        FormatDetailHandler dh = getEntryDetails(puid);

        BitstreamFormat bsf = FormatRegistryManager.createBitstreamFormat(context);
        doUpdateBitstreamFormat(bsf, puid, dh);
         
        // XXX should be able to consult a policy engine here, someday?
        bsf.setSupportLevel(getDefaultSupportLevel());
         
        return bsf;
    }

    /**
     * Re-import the file format metadata from the PRONOM registry.
     *
     * @param context the database context
     * @param bsf the BitstreamFormat to update
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @throws FormatRegistryException for any failure contacting or accessing the registry, or if the identifier cannot be found.
     */
    public void updateBitstreamFormat(Context context, BitstreamFormat bsf, String puid)
        throws FormatRegistryException, AuthorizeException, SQLException
    {
        init();
        FormatDetailHandler dh = getEntryDetails(puid);

        // check for any "new" identifier collisions
        for (String nsid : dh.identifiers)
        {
            BitstreamFormat other = BitstreamFormat.findByIdentifier(context,
                                                                     FormatRegistryManager.namespaceOf(nsid),
                                                                     FormatRegistryManager.identifierOf(nsid), false);
            if (other != null && !other.equals(bsf))
               throw new FormatRegistryException("Cannot update "+bsf+
                 " from PRONOM PUID="+puid+", since the identifier "+nsid+
                 " it references is already bound to a different format: "+other);
        }
        doUpdateBitstreamFormat(bsf, puid, dh);
    }

    /**
     * Get the time when the given registry entry was last changed.
     * PRONOM has two possible dates in the record, although they are
     * not available consistently.
     *
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return date/time when this entry was last changed, or null if not available.
     * @throws FormatRegistryException upon failure accessing the registry.
     */
    public Date getLastModified(String puid)
        throws FormatRegistryException
    {
        init();

        FormatDetailHandler dh = getEntryDetails(puid);
        if (dh.lastUpdatedDate != null)
            return dh.lastUpdatedDate;
        else if (dh.provenanceSourceDate != null)
            return dh.provenanceSourceDate;
        else
            return null;
    }

    /**
     * Predicate, true if format named by sub is a subtype or
     * otherwise "conforms" to the format defined by fmt.
     *
     * For PRONOM, look at RelatedFormat elements in detailed description;
     * define "X conformsTo Y" as any of:
     *    X  Is subtype of  Y
     *    X  Has priority over  Y  (this means they share internal signatures)
     *    Y  Is supertype of  X
     *    Y  Has lower priority than  X
     *
     * @param sub bare (NON-namespaced) external identifier of the subtype format
     * @param fmt bare (NON-namespaced) external identifier of the supertype format
     * @throws FormatRegistryException upon failure accessing the registry.
     */
    public boolean conformsTo(String sub, String fmt)
        throws FormatRegistryException
    {
        init();

        // the easy case?
        if (sub.equals(fmt))
            return true;

        FormatRelationshipHandler subr = getEntryRelationships(sub);
        FormatRelationshipHandler fmtr = getEntryRelationships(fmt);

        if (log.isDebugEnabled())
            log.debug("conformsTo: Checking relationships: subtype="+sub+", format="+fmt);

        String subid = puid2formatID.get(sub);
        String fmtid = puid2formatID.get(fmt);
        if (subid == null)
            throw new FormatRegistryNotFoundException("Cannot find formatID for PUID=\""+sub+"\", need to update the index?");
        if (fmtid == null)
            throw new FormatRegistryNotFoundException("Cannot find formatID for PUID=\""+fmt+"\", need to update the index?");
        for (RelEntry e : subr.relationships)
        {
            if (e.key.equals(fmtid) &&
                (e.value.equals("Is subtype of") ||
                 e.value.equals("Has priority over")))
            {
                log.debug("conformsTo: Found subtype is-sub/has-prio over format");
                return true;
            }
        }
        for (RelEntry e : fmtr.relationships)
        {
            if (e.key.equals(subid) &&
                (e.value.equals("Is supertype of") ||
                 e.value.equals("Has lower priority than")))
            {
                log.debug("conformsTo: Found format is-super/has-lower-prio than sub");
                return true;
            }
        }
        return false;
    }

    /**
     * Free any resources associated with this registry connection,
     * since it will not be used any more.
     */
    public void shutdown()
    {
    }

    /**
     * Administrative command-line UI: decache (update) the local cached
     * copies of PRONOM documents on demand.
     *
     * Usage:  java PRONOMFormatRegistry  -u  [ puid | -a ]
     */
    public static void main(String argv[])
    {
        String Usage = "java "+PRONOMFormatRegistry.class.getName()+" -u  [ puid | -a ]";
        try
        {
            boolean uArg = false;
            boolean all = false;
            String puid = null;
            for (String arg : argv)
            {
                if (arg.equals("-u"))
                    uArg = true;
                else if (arg.equals("-a"))
                    all = true;
                else if (!arg.startsWith("-"))
                    puid = arg;
                else
                {
                    System.err.println("ERROR, unrecognized option: "+arg+"\n"+Usage);
                    System.exit(1);
                }
            }
            if (uArg)
            {
                // Reload index of all formats
                if (all)
                {
                    File path = getIndexPath();
                    new PRONOMFormatRegistry().decacheEntryList(path);
                    System.err.println("Reloaded PRONOM format index in "+path);
                }
                else if (puid != null)
                {
                    System.err.println("Reloading of a PUID is not implemented yet.");
                }
                else
                {
                    System.err.println("Missing option or argument.\n"+Usage);
                }
            }
        }
        catch (Exception e)
        {
            System.err.println("Got exception: "+e);
            e.printStackTrace();
        }
    }

    // Set the fields of a BSF from a PRONOM entry; use when importing a
    // new BSF as well as updating since override is off initially.
    private void doUpdateBitstreamFormat(BitstreamFormat bsf, String puid, FormatDetailHandler dh)
        throws FormatRegistryException, AuthorizeException, SQLException
    {
        if (!bsf.isOverrideName())
        {
            String name = dh.formatName;
            if (dh.formatVersion != null && dh.formatVersion.length() > 0)
                name += " "+dh.formatVersion;
            bsf.setName(name);
        }
        if (dh.formatDescription != null && !bsf.isOverrideDescription())
            bsf.setDescription(dh.formatDescription);
        if (dh.canonicalExtension != null && !bsf.isOverrideCanonicalExtension())
            bsf.setCanonicalExtension(dh.canonicalExtension);
        if (!bsf.isOverrideMIMEType() &&
            (bsf.getMIMEType() == null || dh.mimeType != null))
        {
            bsf.setMIMEType((dh.mimeType != null) ? dh.mimeType : FormatRegistryManager.DEFAULT_MIME_TYPE);
        }
        for (String nsid : dh.identifiers)
            bsf.addIdentifier(nsid);

        bsf.setLastImported(getNamespace(), puid, new Date());
    }

    /**
     * Get the DEFAULT support level applied to a new bitstream format from
     * the DSpace configuration.  It must be one of
     * <code>UNKNOWN</code>,<code>KNOWN</code> or <code>SUPPORTED</code>,
     * so translate to the corresponding constant.
     *
     * @return the support level
     */
    private int getDefaultSupportLevel()
    {
        if (defaultSupportLevel < 0)
        {
            String key = "formatRegistry."+getNamespace()+CONFIG_SUPPORTLEVEL_SUFFIX;
            String name = ConfigurationManager.getProperty(key);
            if (name == null)
                defaultSupportLevel = BitstreamFormat.UNKNOWN;
            else
            {
                name = name.trim();
                for (int i = 0; i < BitstreamFormat.supportLevelText.length; ++i)
                {
                    if (name.equalsIgnoreCase(BitstreamFormat.supportLevelText[i]))
                    {
                        defaultSupportLevel = i;
                        break;
                    }
                }
                if (defaultSupportLevel < 0)
                {
                    log.error("BitstreamFormat support level configuration value unrecognized: "+key+"=\""+name+"\"");
                    defaultSupportLevel = BitstreamFormat.UNKNOWN;
                }
            }
        }
        return defaultSupportLevel;
    }

    // return path to the "index" file with list of all formats, that
    // lets us map PUID to XML-specific FormatID.
    private static File getIndexPath()
    {
        return new File(getDataDirectory(), INDEX_FILE);
    }

    // load the PUID-to-FormatID map from cache of "index" file.
    private void init()
        throws FormatRegistryException
    {
        if (puid2formatID == null)
        {
            File path = getIndexPath();
            try
            {
                if (!path.exists())
                    decacheEntryList(path);

                log.info("Loading PRONOM format PUID to XML-ID map from path="+path);

                // FormatListHandler fills out puid2formatID map as side-effect.
                puid2formatID = new HashMap<String,String>();
                startParse(new FileInputStream(path), new FormatListHandler());
            }
            catch (IOException e)
            {
                log.error("Failed parsing format list in "+path, e);
                throw new IllegalStateException("Failed parsing format list in "+path, e);
            }
            catch (SAXException e)
            {
                log.error("Failed parsing format list in "+path, e);
                throw new IllegalStateException("Failed parsing format list in "+path, e);
            }
        }
    }

    // get cached copy of "index" file (PRONOM entry list in XML) from PRONOM website.
    // XXX NOTE: This is currently implemented as a faked POST request from the
    // interactive web UI, because the RESTful PRONOM PUID resolver doesn't
    // give access to the index.  We _need_ this index because format
    // relationships (subclass, etc) are expressed through FormatID.
    private void decacheEntryList(File path)
        throws FormatRegistryException, IOException
    {
        // generate faked UI request:
        URL cu = getContactURL();
        if (cu == null)
            throw new IOException("Configuration is missing contact URL for PRONOM registry.");
        String url = cu.toString()+"/PRONOM/Format/proFormatListAction.aspx";
        PostMethod post = new PostMethod(url);
        NameValuePair[] postArgs = {
          new NameValuePair("strAction", "Save As XML"),
          new NameValuePair("strFileFormatName", "%%"),
          new NameValuePair("strLastQueryStatus", "formatname"),
          new NameValuePair("strOrderBy", "order_extension"),
          new NameValuePair("strLastQueryFormatName", "%%")
        };
        post.setRequestBody(postArgs);

        log.info("Loading PRONOM entry list from \""+url+"\"");
        try
        {
            HttpClient hc = new HttpClient();
            int status = hc.executeMethod(post);
            if (status != HttpStatus.SC_OK)
            {
                log.error("Failed getting PRONOM format list returns failed status = "+String.valueOf(status));
                throw new IOException("Failed getting PRONOM format list returns failed status = "+String.valueOf(status)+
                            ", reason="+post.getStatusLine().getReasonPhrase());
            }
            OutputStream out = new FileOutputStream(path);
            Utils.copy(post.getResponseBodyAsStream(), out);
            out.close();
            log.debug("Created local cache of PRONOM format list in "+path);
        }
        finally
        {
            post.releaseConnection();
        }
    }

    // Return the path to directory where local file cache of PRONOM XML stored.
    // Default is "${dspace.dir}/config/formats/PRONOM/"
    private static File getDataDirectory()
    {
        if (dataDirectory == null)
        {
            String key = "formatRegistry."+NAMESPACE+CONFIG_DIRECTORY_SUFFIX;
            String dir = ConfigurationManager.getProperty(key);
            if (dir == null)
            {
                dir = ConfigurationManager.getProperty("dspace.dir");
                if (dir != null)
                    dataDirectory = new File(dir.trim() + File.separator + "config" +
                        File.separator + "formats" +
                        File.separator + "PRONOM" + File.separator);
                else
                    log.error("Cannot determine PRONOM data directory, no config for "+key+", or dspace.dir");
            }
            else
            {
                dir = dir.trim();
                if (!dir.endsWith(File.separator))
                    dir += File.separator;
                dataDirectory = new File(dir);
            }
        }
        return dataDirectory;
    }


    // get the Format Detail (technical MD) data from a PUID entry
    // ONLY return a valid handler instance, otherwise throw an exception.
    private FormatDetailHandler getEntryDetails(String puid)
        throws FormatRegistryException
    {
        String id = puid2formatID.get(puid);
        if (id == null)
            throw new FormatRegistryNotFoundException("Cannot find formatID for PUID=\""+puid+"\", need to update the index?");

        FormatDetailHandler dhandler = (FormatDetailHandler)getParsedEntryInternal(puid, new FormatDetailHandler());
        if (dhandler.formatID == null ||
            !id.equals(dhandler.formatID) ||
            dhandler.formatName == null)
            throw new FormatRegistryException("Invalid PRONOM detail record results for Format PUID="+puid);
        if (log.isDebugEnabled())
            log.debug("Got details for PUID="+puid+": "+dhandler.toString());
        return dhandler;
    }

    // get the Format Relationship data from a PUID entry
    // ONLY allow non-null to be returned, otherwise throw.
    // XXX may want to cache (memoize) this method for performance,
    //  since it gets called a lot by conformsTo().
    private FormatRelationshipHandler getEntryRelationships(String puid)
        throws FormatRegistryException
    {
        FormatRelationshipHandler result = (FormatRelationshipHandler)
            getParsedEntryInternal(puid, new FormatRelationshipHandler());
        if (result == null)
            throw new FormatRegistryException("Failed getting format entry from PRONOM, for PUID="+puid);
        return result;
    }

    // driver to parse a PUID's entry and return the populated handler.
    private PronomHandler getParsedEntryInternal(String puid, PronomHandler hdl)
        throws FormatRegistryException
    {
        String id = puid2formatID.get(puid);

        if (id == null)
            throw new FormatRegistryNotFoundException("Cannot find formatID for PUID=\""+puid+"\", need to update the index?");

        //XXX TEMP
        // log.debug("Reading Details from PRONOM for ID="+id);
        try
        {
            startParse(getFormatEntry(puid), hdl);
            return hdl;
        }
        catch (IOException e)
        {
            log.error("Failed parsing format detail for ID="+id, e);
            throw new IllegalStateException("Failed parsing format detail for ID="+id, e);
        }
        catch (SAXException e)
        {
            log.error("Failed parsing format detail for ID="+id, e);
            throw new IllegalStateException("Failed parsing format detail for ID="+id, e);
        }
    }

    // Get format entry from PRONOM as stream of XML document
    // Look for local cache of this record.
    // If not found, fetch record from PRONOM and save it locally.
    private InputStream getFormatEntry(String puid)
        throws IOException, FormatRegistryException
    {
        // sanitize the PUID for a local filename by converting slash to '_'
        // since slash is the filename separator on Unix.  don't bother
        // checking File.separatorChar, keep it consistent on all platforms.
        String filePUID = puid.replace('/', '_');
        File formatFile = new File(getDataDirectory(), "details_"+filePUID+".xml");
        if (!formatFile.exists())
        {
            URL cu = getContactURL();
            if (cu == null)
                throw new IOException("Configuration is missing contact URL for PRONOM registry.");
            String url = cu.toString()+"/pronom/"+puid+".xml";
            log.debug("Decaching PUID="+puid+", URL=\""+url+"\" into file="+formatFile);
            GetMethod get = new GetMethod(url);

            // NOTE: has no effect on stupid IIS server but protocol is correct.
            get.addRequestHeader(new Header("Accept", "text/xml"));
            try
            {
                HttpClient hc = new HttpClient();
                int status = hc.executeMethod(get);
                if (status != HttpStatus.SC_OK)
                {
                    log.error("Getting detail for PUID="+puid+" returns failed status = "+String.valueOf(status));
                    throw new FormatRegistryNotFoundException("Failed to retrieve network resource for PRONOM Format Detail PUID="+
                        puid+",  HTTP status="+String.valueOf(status));
                }
                else
                {
                    Header ct = get.getResponseHeader("Content-Type");
                    if (ct == null)
                        throw new FormatRegistryNotFoundException("Failed to retrieve network resource for PRONOM Format Detail PUID="+
                          puid+",  malformed HTTP response.");
                    if (ct.toExternalForm().indexOf("text/xml") < 0)
                        throw new FormatRegistryNotFoundException("Failed to retrieve network resource for PRONOM Format Detail PUID="+
                          puid+",  returned unacceptable format in HTTP response (probably not found).");
                }
                OutputStream out = new FileOutputStream(formatFile);
                Utils.copy(get.getResponseBodyAsStream(), out);
                out.close();
            }
            finally
            {
                get.releaseConnection();
            }
        }
        return new FileInputStream(formatFile);
    }

    // launch SAX parser with a given handler
    private static void startParse(InputStream in, PronomHandler handler)
        throws SAXException, IOException
    {
        try
        {
            SAXParserFactory pf = SAXParserFactory.newInstance();
            SAXParser sp = pf.newSAXParser();
            XMLReader pkgReader = sp.getXMLReader();

            // XXX FIXME: should turn off validation here explicitly, but
            //  it seems to be off by default.

            pkgReader.setFeature("http://xml.org/sax/features/namespaces", true);
            pkgReader.setContentHandler(handler);
            pkgReader.setErrorHandler(handler);
            pkgReader.parse(new InputSource(in));
        }
        catch (ParserConfigurationException pe)
        {
            throw new SAXException(pe);
        }
    }

    /**
     * ========================================================
     *  SAX handlers for parsing the bits we need out of PRONOM files
     */

    // Shared handler framework for all other handlers;
    //get text values, handle errors.
    static abstract class PronomHandler
        extends DefaultHandler
    {
        protected String value = null;

        // NOTE:  text value MAY be presented in multiple calls, even if
        // it all one word, so be ready to splice it together.
        // BEWARE:  subclass's startElement method should call super()
        // to null out 'value'.  (Don't you miss the method combination
        // options of a real object system like CLOS?)
        public void characters(char[] ch, int start, int length)
            throws SAXException
        {
            String newValue = new String(ch, start, length);
            if (newValue.length() > 0)
            {
                if (value == null)
                    value = newValue;
                else
                    value += newValue;
            }
        }

        // subclass overriding this MUST call it with super()
        public void startElement(String namespaceURI, String localName,
                                 String qName, Attributes atts)
            throws SAXException
        {
            value = null;
        }

        public void error(SAXParseException exception)
            throws SAXException
        {
            throw new SAXException(exception);
        }

        public void fatalError(SAXParseException exception)
            throws SAXException
        {
            throw new SAXException(exception);
        }
    }

    /**
     * Read format list and fill out the puid2formatID map.
     */
    static class FormatListHandler
        extends PronomHandler
    {
        private String formatid;
        private String puid;

        public List<String> results = new ArrayList<String>();

        // reset data collection on the way in just in case.
        // ALSO null out value for superclass.
        public void startElement(String namespaceURI, String localName,
                                 String qName, Attributes atts)
            throws SAXException
        {
            if (localName.equals("FileType"))
            {
                formatid = null;
                puid = null;
            }
            super.startElement(namespaceURI,localName,qName,atts);
        }

        // save contents of FormatID and PUID to put them in a map
        // entry when FileType is closed.
        public void endElement(String namespaceURI, String localName,
                                 String qName)
            throws SAXException
        {
            if (localName.equals("FormatID"))
                formatid = value;
            else if (localName.equals("PUID"))
                puid = value;

            // on the way out of a FileType, log the ID
            else if (localName.equals("FileType"))
            {
                if (formatid == null || puid == null)
                    log.warn(String.format("FormatListHandler: Got bad FileType field, missing field, FormatID=%s, PUID=%s", formatid, puid));
                else
                {
                    puid2formatID.put(puid, formatid);
                    if (log.isDebugEnabled())
                        log.debug("Adding puid2formatID entry, \""+puid+"\" -> \""+formatid+"\"");
                }
                formatid = null;
                puid = null;
            }
        }
    }

    /**
     * Get useful details from a format entry:
     *  formatID - XML ID (numeric) used as reference in List file, relations.
     *  formatName - short human-readable name
     *  formatVersion - version suffix for name, if any.
     *  formatDescription - longer description
     *  mimeType - MIME designation, e.g. text/plain
     *  canonicalExtension - a filename extension to assign to files of this type
     *  provenanceSourceDate - original date?
     *  lastUpdatedDate - date when last modified ?
     *  identifiers - all PUID external identifiers
     */
    static class FormatDetailHandler
        extends PronomHandler
    {
        private boolean inFileFormatIdentifier = false;
        private boolean inExternalSignature = false;
        private String identifier = null;
        private String identifierType = null;
        public String formatID = null;
        public String formatName = null;
        public String formatVersion = null;
        public String formatDescription = null;
        public String mimeType = null;
        public String canonicalExtension = null;
        public Date provenanceSourceDate = null;
        public Date lastUpdatedDate = null;
        public List<String> identifiers = new ArrayList<String>();

        public void startElement(String namespaceURI, String localName,
                                 String qName, Attributes atts)
            throws SAXException
        {
            //XXX TEMP
            //log.debug("FormatDetailHandler: Starting "+localName);

            // at start of FileFormatIdentifier, clear the record:
            if (localName.equals("FileFormatIdentifier"))
                inFileFormatIdentifier = true;
            else if (localName.equals("ExternalSignature"))
                inExternalSignature = true;
            super.startElement(namespaceURI,localName,qName,atts);
        }

        public void endElement(String namespaceURI, String localName,
                                 String qName)
            throws SAXException
        {
            //XXX TEMP
            //log.debug("FormatDetailHandler: Ending "+localName);

            // at end of FileFormatIdentifier, enter the record:
            if (localName.equals("FileFormatIdentifier"))
            {
                if (identifierType != null && identifier != null)
                {
                    if (identifierType.equals("MIME"))
                        mimeType = identifier;
                    else if (identifierType.equals("PUID"))
                        identifiers.add(FormatRegistryManager.makeIdentifier(NAMESPACE, identifier));
                }
                inFileFormatIdentifier = false;
                identifier = null;
                identifierType = null;
            }
            else if (localName.equals("ExternalSignature"))
                inExternalSignature = false;
            else if (localName.equals("Signature") && inExternalSignature &&
                     canonicalExtension == null)
                canonicalExtension = value.trim();
            else if (localName.equals("Identifier"))
                identifier = value.trim();
            else if (localName.equals("IdentifierType"))
                identifierType = value.trim();
            else if (localName.equals("FormatName"))
                formatName = value.trim();
            else if (localName.equals("FormatVersion"))
                formatVersion = value.trim();
            else if (localName.equals("FormatDescription"))
                formatDescription = value;
            else if (localName.equals("LastUpdatedDate") && value != null)
                lastUpdatedDate = parsePronomDate(value);
            else if (localName.equals("ProvenanceSourceDate") && value != null)
                provenanceSourceDate = parsePronomDate(value);
            else if (localName.equals("FormatID"))
                formatID = value.trim();
        }

        public String toString()
        {
            return String.format("<FormatDetailHandler ID=%s, formatName=%s, formatVersion=%s, formatDescription=%s, mimeType=%s, canonExt=%s lastUpdate=%s, ProvSrcDate=%s />",
                                 formatID,formatName,formatVersion,formatDescription,
                                 mimeType,canonicalExtension,
                                 lastUpdatedDate,provenanceSourceDate);
        }

        // parse the "22 Aug 2006" dates used in PRONOM XML dumps.
        private static SimpleDateFormat pronomDateFmt =
            new SimpleDateFormat("d MMM yyyy");

        private static Date parsePronomDate(String s)
        {
            try
            {
                //XXX TEMP
                //log.debug("Parsing date=\""+s+"\" with pattern="+pronomDateFmt.toPattern()+", example="+pronomDateFmt.format(new Date()));
                pronomDateFmt.setLenient(true);
                return pronomDateFmt.parse(s);
            }
            catch (ParseException e)
            {
                log.warn("Got ParseException in entry date: "+ e);
            }
            return null;
        }
    }

    /**
     * Like a Map.Entry for storing format relationships, but since
     * we may have duplicate keys these must be stored on a list
     * instead of a map.
     */
    static class RelEntry
    {
        String key;
        String value;
        RelEntry(String k, String v)
        {
            key = k;
            value = v;
        }
    }

    /**
     * just extracts the "relationship" keys from a format-detail XML document
     * and builds up a List of mappings from ID to relationship-type,
     * where ID is the XML ID of the othe format, not PUID.
     * This is derived from the elements:
     *    RelatedFormat
     *      RelatedFormatID value==ID
     *      RelationshipType value== one of:
     *              "Is supertype of"
     *              "Is subtype of"
     *              "Has lower priority than"
     *              "Has priority over"
     */
    static class FormatRelationshipHandler
        extends PronomHandler
    {
        // results of parse
        protected List<RelEntry> relationships = new ArrayList<RelEntry>();

        private boolean inRelatedFormat = false;
        private String formatID = null;
        private String relType = null;

        public void startElement(String namespaceURI, String localName,
                                 String qName, Attributes atts)
            throws SAXException
        {
            //XXX TEMP
            //log.debug("FormatRelationshipHandler: Starting "+localName);

            // at start of FileFormatIdentifier, clear the record:
            if (localName.equals("RelatedFormat"))
                inRelatedFormat = true;
            super.startElement(namespaceURI,localName,qName,atts);
        }

        public void endElement(String namespaceURI, String localName,
                                 String qName)
            throws SAXException
        {
            //XXX TEMP
            //log.debug("FormatDetailHandler: Ending "+localName);

            // at end of RelatedFormat enter the record:
            if (localName.equals("RelatedFormat"))
            {
                if (formatID != null && relType != null)
                {
                    relationships.add(
                        new RelEntry(formatID,relType));
                }
                else
                    log.error("Got RelatedFormat missing one/both of RelatedFormatID, RelationshipType!");
                inRelatedFormat = false;
                formatID = null;
                relType = null;
            }
            else if (localName.equals("RelatedFormatID"))
                formatID = value;
            else if (localName.equals("RelationshipType"))
                relType = value;
        }

        public String toString()
        {
            String ls = "";
            for (RelEntry e : relationships)
            {
                ls += "<Entry k=\""+e.key+"\", v=\""+e.value+"\">, ";
            }
            return String.format("<FormatRelationshipHandler result=(%s) />",
                                 ls);
        }
    }
}
