package org.dspace.content.format;

import org.dspace.content.Bitstream;
import org.dspace.core.ConfigurationManager;
import org.dspace.core.Context;
import org.dspace.authorize.AuthorizeException;

import java.io.IOException;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;

/**
 * Text Subtype File Format Identifier
 *
 * This identifier method mediates between the heuristic-based text file
 * identifier (TextHeuristicIdentifier) and subtypes of plain text which
 * are difficult to identify by internal characteristics but which have
 * external signatures such as filename extensions.
 *
 * When this method detects both a heuristically-identified text type,
 * and a subtype identified at a lower confidence, it "refines" the
 * format identification to the subtype:  it elevates the subtype to the
 * highest confidence between the two and places it first.
 *
 * For example, HTML CSS files are plain text, but virtually impossible
 * to identify by internal signature.  If this method finds a
 * Circumstantial CSS hit in the results, along with a Heuristic ASCII
 * text hit (the supertype), then it promotes the CSS hit to the
 * Heuristic confidence and moves it to the head of the results list.
 *
 * Configuration:
 *   formatIdentifier.TextSubtypeHelper.subtypes = id, id, ...
 *
 * <p>
 * @author Larry Stone
 * @see TextHeuristicIdentifier
 * @version $Revision: 1323 $
 */
public class TextSubtypeHelper
    extends TextHeuristicIdentifier
{
    /** log4j logger */
    private static Logger log = Logger.getLogger(TextSubtypeHelper.class);

    // Namespaced identifiers of all Text subtypes to be considered for
    // promotion.  The list comes from the DSpace configuration.
    private static Set<String> subtypes = null;

    // initial setup -- load configured subtype list.
    private void init()
    {
        if (subtypes == null)
        {
            String key = "formatIdentifier.TextSubtypeHelper.subtypes";
            subtypes = new HashSet<String>();
            String rawNsids = ConfigurationManager.getProperty(key);
            if (rawNsids == null)
                log.error("Fatal Configuration error: No value for "+key);
            else
                subtypes.addAll(Arrays.asList(rawNsids.trim().split("\\s*,\\s*")));
        }
    }

    /**
     * Identifier method - which actually just refines existing results.
     * <p>
     * @param context the DB context
     * @param target the Bitstream to identify
     * @param results List of results produced by previous identifyFormat methods
     * @return list of results, possibly modified or replacing the result param
     * @throws FormatIdentifierException when a fatal error in the process
     *   of identifying the format prevents it from returning conclusive results
     */
    public List<FormatHit> identifyFormat(Context context, Bitstream target, List<FormatHit> results)
        throws FormatIdentifierException, AuthorizeException, SQLException, IOException
    {
        init();

        // if leading result is a Text format and it was heuristically
        // identified, look for lower-confidence hit of a subtype to promote..
        if (!results.isEmpty())
        {
            FormatHit first = results.get(0);
            if (getTextFormatIdentifiers().contains(first.getIdentifier()) &&
                first.confidence.compareTo(FormatConfidence.HEURISTIC) >= 0)
            {
                for (FormatHit r : results)
                {
                    // helpful for debugging, see what hits are evaluated
                    if (log.isDebugEnabled())
                        log.debug("Checking for subtype, hit format="+r.getIdentifier());

                    if (r != first &&
                        r.confidence.compareTo(FormatConfidence.HEURISTIC) < 0 &&
                        subtypes.contains(r.getIdentifier()))
                    {
                        results.remove(r);
                        r.confidence = first.confidence;
                        results.add(0, r);
                        if (log.isDebugEnabled())
                            log.debug("Promoting hit for "+r.getIdentifier());
                        return results;
                    }
                }
            }
        }
        return results;
    }
}
